# This module is derived from the train.py module in the following repository:
# https://github.com/NVlabs/FourCastNet. The corresponding license is
# provided below.

# BSD 3-Clause License
#
# Copyright (c) 2022, FourCastNet authors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The code was authored by the following people:
#
# Jaideep Pathak - NVIDIA Corporation
# Shashank Subramanian - NERSC, Lawrence Berkeley National Laboratory
# Peter Harrington - NERSC, Lawrence Berkeley National Laboratory
# Sanjeev Raja - NERSC, Lawrence Berkeley National Laboratory
# Ashesh Chattopadhyay - Rice University
# Morteza Mardani - NVIDIA Corporation
# Thorsten Kurth - NVIDIA Corporation
# David Hall - NVIDIA Corporation
# Zongyi Li - California Institute of Technology, NVIDIA Corporation
# Kamyar Azizzadenesheli - Purdue University
# Pedram Hassanzadeh - Rice University
# Karthik Kashinath - NVIDIA Corporation
# Animashree Anandkumar - California Institute of Technology, NVIDIA Corporation

import abc
import contextlib
import gc
import logging
import os
import time
import uuid
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    List,
    Optional,
    Protocol,
    TypeVar,
)

import torch

import fme
from fme.core.distributed import Distributed
from fme.core.ema import EMATracker
from fme.core.generics.aggregator import AggregatorABC, InferenceAggregatorABC
from fme.core.generics.data import GriddedDataABC, InferenceDataABC
from fme.core.generics.inference import run_inference
from fme.core.generics.train_stepper import TrainOutputABC, TrainStepperABC
from fme.core.optimization import NullOptimization, Optimization
from fme.core.timing import GlobalTimer
from fme.core.typing_ import Slice
from fme.core.wandb import WandB


class EndOfBatchCallback(Protocol):
    def __call__(self) -> None: ...


class TrainConfigProtocol(Protocol):
    __dataclass_fields__: ClassVar[Dict[str, Any]]

    @property
    def experiment_dir(self) -> str: ...

    @property
    def checkpoint_dir(self) -> str: ...

    @property
    def max_epochs(self) -> int: ...

    @property
    def save_checkpoint(self) -> bool: ...

    @property
    def validate_using_ema(self) -> bool: ...

    @property
    def log_train_every_n_batches(self) -> int: ...

    @property
    def segment_epochs(self) -> Optional[int]: ...

    @property
    def checkpoint_save_epochs(self) -> Optional[Slice]: ...

    @property
    def ema_checkpoint_save_epochs(self) -> Optional[Slice]: ...

    def clean_wandb(self, experiment_dir: str) -> None: ...

    def get_inference_epochs(self) -> List[int]: ...


PS = TypeVar("PS", contravariant=True)  # prognostic state
TO = TypeVar("TO", bound="TrainOutputABC")  # train output
BD = TypeVar("BD")  # batch data for training
FD = TypeVar("FD")  # forcing data for inference
SD = TypeVar("SD")  # stepped data from inference


class AggregatorBuilderABC(abc.ABC, Generic[PS, TO, SD]):
    @abc.abstractmethod
    def get_train_aggregator(self) -> AggregatorABC[TO]:
        pass

    @abc.abstractmethod
    def get_validation_aggregator(self) -> AggregatorABC[TO]:
        pass

    @abc.abstractmethod
    def get_inference_aggregator(self) -> InferenceAggregatorABC[PS, SD]:
        pass


class CheckpointPaths:
    def __init__(self, checkpoint_dir: str):
        self.checkpoint_dir = checkpoint_dir

    @property
    def latest_checkpoint_path(self) -> str:
        return os.path.join(self.checkpoint_dir, "ckpt.tar")

    @property
    def best_checkpoint_path(self) -> str:
        return os.path.join(self.checkpoint_dir, "best_ckpt.tar")

    @property
    def best_inference_checkpoint_path(self) -> str:
        return os.path.join(self.checkpoint_dir, "best_inference_ckpt.tar")

    @property
    def ema_checkpoint_path(self) -> str:
        return os.path.join(self.checkpoint_dir, "ema_ckpt.tar")

    def epoch_checkpoint_path(self, epoch: int) -> str:
        return os.path.join(self.checkpoint_dir, f"ckpt_{epoch:04d}.tar")

    def ema_epoch_checkpoint_path(self, epoch: int) -> str:
        return os.path.join(self.checkpoint_dir, f"ema_ckpt_{epoch:04d}.tar")


class Trainer:
    def __init__(
        self,
        train_data: GriddedDataABC[BD],
        validation_data: GriddedDataABC[BD],
        inference_data: InferenceDataABC[PS, FD],
        stepper: TrainStepperABC[PS, BD, FD, SD, TO],
        build_optimization: Callable[[torch.nn.ModuleList], Optimization],
        build_ema: Callable[[torch.nn.ModuleList], EMATracker],
        config: TrainConfigProtocol,
        aggregator_builder: AggregatorBuilderABC[PS, TO, SD],
        end_of_batch_callback: EndOfBatchCallback = lambda: None,
        do_gc_collect: bool = True,
    ):
        logging.info(f"Current device is {fme.get_device()}")
        self.dist = Distributed.get_instance()
        if self.dist.is_root():
            if not os.path.isdir(config.experiment_dir):
                os.makedirs(config.experiment_dir)
            if not os.path.isdir(config.checkpoint_dir):
                os.makedirs(config.checkpoint_dir)
        self.config = config
        self.paths = CheckpointPaths(config.checkpoint_dir)

        logging.info("rank %d, begin data loader init" % self.dist.rank)
        self.train_data = train_data
        self.valid_data = validation_data
        logging.info("rank %d, data loader initialized" % self.dist.rank)
        for gridded_data, name in zip(
            (self.train_data, self.valid_data), ("train", "valid")
        ):
            gridded_data.log_info(name)

        self.num_batches_seen = 0
        self._start_epoch = 0
        self._model_epoch = self._start_epoch
        self.num_batches_seen = 0
        self._best_validation_loss = torch.inf
        self._best_inference_error = torch.inf

        self.stepper = stepper
        self.optimization = build_optimization(stepper.modules)
        self._end_of_batch_ops = end_of_batch_callback
        self._no_optimization = NullOptimization()
        self._aggregator_builder = aggregator_builder

        resuming = os.path.isfile(self.paths.latest_checkpoint_path)
        if resuming:
            logging.info(f"Resuming training from {self.paths.latest_checkpoint_path}")
            self.restore_checkpoint(
                self.paths.latest_checkpoint_path, self.paths.ema_checkpoint_path
            )

        wandb = WandB.get_instance()
        wandb.watch(self.stepper.modules)

        logging.info(
            (
                "Number of trainable model parameters: "
                f"{count_parameters(self.stepper.modules)}"
            )
        )

        self._inference_data = inference_data
        self._ema = build_ema(stepper.modules)
        self._do_gc_collect = do_gc_collect

    def switch_off_grad(self, model: torch.nn.Module):
        for param in model.parameters():
            param.requires_grad = False

    def train(self):
        logging.info("Starting Training Loop...")

        self._model_epoch = self._start_epoch
        inference_epochs = self.config.get_inference_epochs()
        if self.config.segment_epochs is None:
            segment_max_epochs = self.config.max_epochs
        else:
            segment_max_epochs = min(
                self._start_epoch + self.config.segment_epochs, self.config.max_epochs
            )
        # "epoch" describes the loop, self._model_epoch describes model weights
        # needed so we can describe the loop even after weights are updated
        for epoch in range(self._start_epoch, segment_max_epochs):
            if self._do_gc_collect:
                # garbage collect to avoid CUDA error in some contexts
                # https://github.com/pytorch/pytorch/issues/67978#issuecomment-1661986812  # noqa: E501
                gc.collect()
            logging.info(f"Epoch: {epoch+1}")
            self.train_data.set_epoch(epoch)

            start_time = time.time()
            logging.info(f"Starting training step on epoch {epoch + 1}")
            train_logs = self.train_one_epoch()
            train_end = time.time()
            logging.info(f"Starting validation step on epoch {epoch + 1}")
            valid_logs = self.validate_one_epoch()
            valid_end = time.time()
            if epoch in inference_epochs:
                logging.info(f"Starting inference step on epoch {epoch + 1}")
                inference_logs = self.inference_one_epoch()
                inference_end: Optional[float] = time.time()
            else:
                inference_logs = {}
                inference_end = None

            train_loss = train_logs["train/mean/loss"]
            valid_loss = valid_logs["val/mean/loss"]
            inference_error = inference_logs.get(
                "inference/time_mean_norm/rmse/channel_mean", None
            )
            # need to get the learning rate before stepping the scheduler
            lr = self.optimization.learning_rate
            self.optimization.step_scheduler(valid_loss)

            if self.dist.is_root():
                if self.config.save_checkpoint:
                    logging.info(f"Saving checkpoints for epoch {epoch + 1}")
                    self.save_all_checkpoints(valid_loss, inference_error)

            time_elapsed = time.time() - start_time
            logging.info(f"Time taken for epoch {epoch + 1} is {time_elapsed} sec")
            logging.info(f"Train loss: {train_loss}. Valid loss: {valid_loss}")
            if inference_error is not None:
                logging.info(f"Inference error: {inference_error}")

            logging.info("Logging to wandb")
            all_logs = {
                **train_logs,
                **valid_logs,
                **inference_logs,
                **{
                    "lr": lr,
                    "epoch": epoch,
                    "epoch_train_seconds": train_end - start_time,
                    "epoch_validation_seconds": valid_end - train_end,
                    "epoch_total_seconds": time_elapsed,
                },
            }
            if inference_end is not None:
                all_logs["epoch_inference_seconds"] = inference_end - valid_end
            wandb = WandB.get_instance()
            wandb.log(all_logs, step=self.num_batches_seen)
        if segment_max_epochs == self.config.max_epochs:
            self.config.clean_wandb(experiment_dir=self.config.experiment_dir)

    def train_one_epoch(self):
        """Train for one epoch and return logs from TrainAggregator."""
        wandb = WandB.get_instance()
        aggregator = self._aggregator_builder.get_train_aggregator()
        n_samples_seen_since_logging = 0
        if self.num_batches_seen == 0:
            # Before training, log the loss on the first batch.
            with torch.no_grad(), GlobalTimer():
                batch = next(iter(self.train_data.loader))
                stepped = self.stepper.train_on_batch(
                    batch,
                    optimization=self._no_optimization,
                )

                if self.config.log_train_every_n_batches > 0:
                    with torch.no_grad():
                        metrics = {
                            f"batch_{name}": self.dist.reduce_mean(metric)
                            for name, metric in sorted(stepped.get_metrics().items())
                        }
                    wandb.log(metrics, step=self.num_batches_seen)
        current_time = time.time()
        for batch in self.train_data.loader:
            with GlobalTimer():
                stepped = self.stepper.train_on_batch(batch, self.optimization)
            aggregator.record_batch(stepped)
            self._end_of_batch_ops()
            self._ema(model=self.stepper.modules)
            self.num_batches_seen += 1
            n_samples_seen_since_logging += self.train_data.batch_size
            if (
                self.config.log_train_every_n_batches > 0
                and self.num_batches_seen % self.config.log_train_every_n_batches == 0
            ):
                with torch.no_grad():
                    metrics = {
                        f"batch_{name}": self.dist.reduce_mean(metric)
                        for name, metric in sorted(stepped.get_metrics().items())
                    }
                duration = time.time() - current_time
                current_time = time.time()
                samples_per_second = n_samples_seen_since_logging / duration
                metrics["training_samples_per_second_on_rank_0"] = samples_per_second
                wandb.log(metrics, step=self.num_batches_seen)
                n_samples_seen_since_logging = 0
        self._model_epoch += 1

        return aggregator.get_logs(label="train")

    @contextlib.contextmanager
    def _validation_context(self):
        """
        The context for running validation.

        In this context, the stepper uses the EMA model if
        `self.config.validate_using_ema` is True.
        """
        if self.config.validate_using_ema:
            with self._ema_context():
                yield
        else:
            yield

    @contextlib.contextmanager
    def _ema_context(self):
        """
        A context where the stepper uses the EMA model.
        """
        self._ema.store(parameters=self.stepper.modules.parameters())
        self._ema.copy_to(model=self.stepper.modules)
        try:
            yield
        finally:
            self._ema.restore(parameters=self.stepper.modules.parameters())

    def validate_one_epoch(self):
        aggregator = self._aggregator_builder.get_validation_aggregator()
        with torch.no_grad(), self._validation_context(), GlobalTimer():
            for batch in self.valid_data.loader:
                stepped = self.stepper.train_on_batch(
                    batch,
                    optimization=NullOptimization(),
                    compute_derived_variables=True,
                )
                aggregator.record_batch(
                    batch=stepped,
                )
        return aggregator.get_logs(label="val")

    def inference_one_epoch(self):
        aggregator = self._aggregator_builder.get_inference_aggregator()
        with torch.no_grad(), self._validation_context(), GlobalTimer():
            run_inference(
                predict=self.stepper.predict_paired,
                data=self._inference_data,
                aggregator=aggregator,
            )
        logs = aggregator.get_summary_logs()
        return {f"inference/{k}": v for k, v in logs.items()}

    def save_checkpoint(self, checkpoint_path):
        # save to a temporary file in case we get pre-empted during save
        temporary_location = os.path.join(
            os.path.dirname(checkpoint_path), f".{uuid.uuid4()}.tmp"
        )
        try:
            torch.save(
                {
                    "num_batches_seen": self.num_batches_seen,
                    "epoch": self._model_epoch,
                    "best_validation_loss": self._best_validation_loss,
                    "best_inference_error": self._best_inference_error,
                    "stepper": self.stepper.get_state(),
                    "optimization": self.optimization.get_state(),
                    "ema": self._ema.get_state(),
                },
                temporary_location,
            )
            os.replace(temporary_location, checkpoint_path)
        finally:
            if os.path.exists(temporary_location):
                os.remove(temporary_location)

    def restore_checkpoint(self, checkpoint_path, ema_checkpoint_path):
        _restore_checkpoint(self, checkpoint_path, ema_checkpoint_path)

    def _epoch_checkpoint_enabled(self, epoch: int) -> bool:
        return epoch_checkpoint_enabled(
            epoch, self.config.max_epochs, self.config.checkpoint_save_epochs
        )

    def _ema_epoch_checkpoint_enabled(self, epoch: int) -> bool:
        return epoch_checkpoint_enabled(
            epoch, self.config.max_epochs, self.config.ema_checkpoint_save_epochs
        )

    def save_all_checkpoints(self, valid_loss: float, inference_error: Optional[float]):
        if self.config.validate_using_ema:
            best_checkpoint_context = self._ema_context
        else:
            best_checkpoint_context = contextlib.nullcontext  # type: ignore
        with best_checkpoint_context():
            save_best_checkpoint = False
            if valid_loss <= self._best_validation_loss:
                logging.info(
                    "Saving lowest validation loss checkpoint to "
                    f"{self.paths.best_checkpoint_path}"
                )
                self._best_validation_loss = valid_loss
                save_best_checkpoint = True  # wait until inference error is updated
            if inference_error is not None and (
                inference_error <= self._best_inference_error
            ):
                logging.info(
                    f"Epoch inference error ({inference_error}) is lower than "
                    f"previous best inference error ({self._best_inference_error})."
                )
                logging.info(
                    "Saving lowest inference error checkpoint to "
                    f"{self.paths.best_inference_checkpoint_path}"
                )
                self._best_inference_error = inference_error
                self.save_checkpoint(self.paths.best_inference_checkpoint_path)
            if save_best_checkpoint:
                self.save_checkpoint(self.paths.best_checkpoint_path)

        logging.info(f"Saving latest checkpoint to {self.paths.latest_checkpoint_path}")
        self.save_checkpoint(self.paths.latest_checkpoint_path)
        with self._ema_context():
            logging.info(
                f"Saving latest EMA checkpoint to {self.paths.ema_checkpoint_path}"
            )
            self.save_checkpoint(self.paths.ema_checkpoint_path)
        if self._epoch_checkpoint_enabled(self._model_epoch):
            epoch_checkpoint_path = self.paths.epoch_checkpoint_path(self._model_epoch)
            logging.info(f"Saving epoch checkpoint to {epoch_checkpoint_path}")
            self.save_checkpoint(epoch_checkpoint_path)
        if self._ema_epoch_checkpoint_enabled(self._model_epoch):
            ema_epoch_checkpoint_path = self.paths.ema_epoch_checkpoint_path(
                self._model_epoch
            )
            logging.info(f"Saving EMA epoch checkpoint to {ema_epoch_checkpoint_path}")
            with self._ema_context():
                self.save_checkpoint(ema_epoch_checkpoint_path)


def _restore_checkpoint(trainer: Trainer, checkpoint_path, ema_checkpoint_path):
    # separated into a function only to make it easier to mock
    checkpoint = torch.load(checkpoint_path, map_location=fme.get_device())
    # restore checkpoint is used for finetuning as well as resuming.
    # If finetuning (i.e., not resuming), restore checkpoint
    # does not load optimizer state, instead uses config specified lr.
    trainer.stepper.load_state(checkpoint["stepper"])
    trainer.optimization.load_state(checkpoint["optimization"])
    trainer.num_batches_seen = checkpoint["num_batches_seen"]
    trainer._start_epoch = checkpoint["epoch"]
    trainer._best_validation_loss = checkpoint["best_validation_loss"]
    trainer._best_inference_error = checkpoint["best_inference_error"]
    ema_checkpoint = torch.load(ema_checkpoint_path, map_location=fme.get_device())
    ema_stepper: TrainStepperABC = type(trainer.stepper).from_state(
        ema_checkpoint["stepper"]
    )
    trainer._ema = EMATracker.from_state(checkpoint["ema"], ema_stepper.modules)


def count_parameters(modules: torch.nn.ModuleList) -> int:
    parameters = 0
    for module in modules:
        for parameter in module.parameters():
            if parameter.requires_grad:
                parameters += parameter.numel()
    return parameters


def epoch_checkpoint_enabled(
    epoch: int, max_epochs: int, save_epochs: Optional[Slice]
) -> bool:
    if save_epochs is None:
        return False
    return epoch in range(max_epochs)[save_epochs.slice]



"""This file contains unit tests of XarrayDataset."""

import dataclasses
import datetime
from collections import namedtuple
from typing import Dict, Iterable, List, Union

import cftime
import numpy as np
import pandas as pd
import pytest
import torch
import xarray as xr

from fme.core.coordinates import LatLonCoordinates
from fme.core.dataset.config import (
    FillNaNsConfig,
    OverwriteConfig,
    TimeSlice,
    XarrayDataConfig,
)
from fme.core.dataset.requirements import DataRequirements
from fme.core.dataset.xarray import (
    XarrayDataset,
    get_cumulative_timesteps,
    get_file_local_index,
    get_raw_times,
    get_timestep,
    get_xarray_dataset,
    repeat_and_increment_time,
)
from fme.core.typing_ import Slice

from .utils import as_broadcasted_tensor, infer_horizontal_dimension_names

SLICE_NONE = slice(None)
MOCK_DATA_FREQ = "3h"
MOCK_DATA_START_DATE = "2003-03"


@dataclasses.dataclass
class VariableNames:
    time_dependent_names: Iterable[str]
    time_invariant_names: Iterable[str]
    initial_condition_names: Iterable[str]

    def _concat(self, *lists):
        return_value = []
        for list in lists:
            return_value.extend(list)
        return return_value

    @property
    def all_names(self) -> List[str]:
        return self._concat(
            self.time_dependent_names,
            self.time_invariant_names,
            self.initial_condition_names,
        )

    @property
    def spatial_resolved_names(self) -> List[str]:
        return self._concat(self.time_dependent_names, self.time_invariant_names)


MockData = namedtuple(
    "MockData", ("tmpdir", "obs_times", "start_times", "start_indices", "var_names")
)


def _get_data(
    tmp_path_factory,
    dirname,
    start,
    end,
    file_freq,
    step_freq,
    calendar,
    with_nans=False,
) -> MockData:
    """Constructs an xarray dataset and saves to disk in netcdf format."""
    obs_times = xr.cftime_range(
        start,
        end,
        freq=step_freq,
        calendar=calendar,
        inclusive="left",
    )
    start_times = xr.cftime_range(
        start,
        end,
        freq=file_freq,
        calendar=calendar,
        inclusive="left",
    )
    obs_delta = obs_times[1] - obs_times[0]
    n_levels = 2
    n_lat, n_lon = 4, 8
    var_names = ["foo", "bar"]
    constant_var = xr.DataArray(
        np.random.randn(n_lat, n_lon).astype(np.float32),
        dims=("lat", "lon"),
    )
    constant_scalar_var = xr.DataArray(1.0).astype(np.float32)
    ak = {f"ak_{i}": float(i) for i in range(n_levels)}
    bk = {f"bk_{i}": float(i + 1) for i in range(n_levels)}
    tmpdir = tmp_path_factory.mktemp(dirname)
    filenames = []
    for i, first in enumerate(start_times):
        if first != start_times[-1]:
            last = start_times[i + 1]
        else:
            last = obs_times[-1] + obs_delta
        time = xr.cftime_range(
            first, last, freq=step_freq, calendar=calendar, inclusive="left"
        )
        data_vars: Dict[str, Union[float, xr.DataArray]] = {**ak, **bk}
        for var_name in var_names:
            data = np.random.randn(len(time), n_lat, n_lon).astype(np.float32)
            if with_nans:
                data[0, :, 0] = np.nan
            data_vars[var_name] = xr.DataArray(data, dims=("time", "lat", "lon"))

        data_varying_scalar = np.random.randn(len(time)).astype(np.float32)
        data_vars["varying_scalar_var"] = xr.DataArray(
            data_varying_scalar, dims=("time",)
        )

        data_vars["constant_var"] = constant_var
        data_vars["constant_scalar_var"] = constant_scalar_var

        coords = {
            "time": xr.DataArray(time, dims=("time",)),
            "lat": xr.DataArray(np.arange(n_lat, dtype=np.float32), dims=("lat",)),
            "lon": xr.DataArray(np.arange(n_lon, dtype=np.float32), dims=("lon",)),
        }

        ds = xr.Dataset(data_vars=data_vars, coords=coords)
        filename = tmpdir / f"{first.strftime('%Y%m%d%H')}.nc"
        ds.to_netcdf(
            filename,
            unlimited_dims=["time"],
            format="NETCDF4",
        )
        filenames.append(filename)

    initial_condition_names = ()
    start_indices = get_cumulative_timesteps(get_raw_times(filenames, "netcdf4"))

    variable_names = VariableNames(
        time_dependent_names=("foo", "bar", "varying_scalar_var"),
        time_invariant_names=("constant_var", "constant_scalar_var"),
        initial_condition_names=initial_condition_names,
    )
    return MockData(tmpdir, obs_times, start_times, start_indices, variable_names)


def get_mock_monthly_netcdfs(tmp_path_factory, dirname, with_nans=False) -> MockData:
    return _get_data(
        tmp_path_factory,
        dirname,
        start=MOCK_DATA_START_DATE,
        end="2003-06",
        file_freq="MS",
        step_freq=MOCK_DATA_FREQ,
        calendar="standard",
        with_nans=with_nans,
    )


@pytest.fixture(scope="session")
def mock_monthly_netcdfs(tmp_path_factory) -> MockData:
    return get_mock_monthly_netcdfs(tmp_path_factory, "month")


@pytest.fixture(scope="session")
def mock_monthly_netcdfs_with_nans(tmp_path_factory) -> MockData:
    return get_mock_monthly_netcdfs(tmp_path_factory, "month_with_nans", with_nans=True)


@pytest.fixture(scope="session")
def mock_monthly_zarr(tmp_path_factory, mock_monthly_netcdfs) -> MockData:
    zarr_parent = tmp_path_factory.mktemp("zarr")
    filename = "data.zarr"
    data = xr.open_mfdataset(
        mock_monthly_netcdfs.tmpdir.glob("*.nc"),
        use_cftime=True,
        data_vars="minimal",
        coords="minimal",
    )
    data.chunk({"time": 240}).to_zarr(zarr_parent / filename)
    return MockData(
        zarr_parent,
        mock_monthly_netcdfs.obs_times,
        mock_monthly_netcdfs.start_times,
        mock_monthly_netcdfs.start_indices,
        mock_monthly_netcdfs.var_names,
    )


@pytest.fixture(scope="session")
def mock_yearly_netcdfs(tmp_path_factory):
    return _get_data(
        tmp_path_factory,
        "yearly",
        start="1999",
        end="2001",
        file_freq="YS",
        step_freq="1D",
        calendar="noleap",
    )


@pytest.mark.parametrize(
    "global_idx,expected_file_local_idx",
    [
        pytest.param(0, (0, 0), id="monthly_file_local_idx_2003_03_01_00"),
        pytest.param(1, (0, 1), id="monthly_file_local_idx_2003_03_01_03"),
        pytest.param(30 * 8, (0, 30 * 8), id="monthly_file_local_idx_2003_03_31_00"),
        pytest.param(31 * 8, (1, 0), id="monthly_file_local_idx_2003_04_01_00"),
        pytest.param(
            (31 + 30 + 20) * 8 - 1,
            (2, 20 * 8 - 1),
            id="monthly_file_local_idx_2003_05_20_21",
        ),
    ],
)
def test_monthly_file_local_index(
    mock_monthly_netcdfs, global_idx, expected_file_local_idx
):
    mock_data: MockData = mock_monthly_netcdfs
    file_local_idx = get_file_local_index(global_idx, mock_data.start_indices)
    assert file_local_idx == expected_file_local_idx
    delta = mock_data.obs_times[1] - mock_data.obs_times[0]
    target_timestamp = np.datetime64(
        cftime.DatetimeGregorian(2003, 3, 1, 0, 0, 0, 0, has_year_zero=False)
        + global_idx * delta
    )
    file_idx, local_idx = file_local_idx
    full_paths = sorted(list(mock_data.tmpdir.glob("*.nc")))
    with xr.open_dataset(full_paths[file_idx], use_cftime=True) as ds:
        assert ds["time"][local_idx].item() == target_timestamp


def _test_monthly_values(
    mock_data: MockData,
    global_idx,
    expected_n_samples=None,
    file_pattern="*.nc",
    engine="netcdf4",
):
    """Runs shape and length checks on the dataset."""
    var_names: VariableNames = mock_data.var_names
    config = XarrayDataConfig(
        data_path=mock_data.tmpdir, file_pattern=file_pattern, engine=engine
    )
    requirements = DataRequirements(names=var_names.all_names, n_timesteps=2)
    dataset = XarrayDataset(config=config, requirements=requirements)
    if expected_n_samples is None:
        expected_n_samples = len(mock_data.obs_times) - 1

    assert len(dataset) == expected_n_samples
    arrays, time = dataset[global_idx]
    with xr.open_mfdataset(
        mock_data.tmpdir.glob(file_pattern),
        engine=engine,
        use_cftime=True,
        data_vars="minimal",
        coords="minimal",
    ) as ds:
        target_times = ds["time"][global_idx : global_idx + 2].drop_vars("time")
        xr.testing.assert_equal(time, target_times)
        lon_dim, lat_dim = infer_horizontal_dimension_names(ds)
        dims = ("time", str(lat_dim), str(lon_dim))
        shape = (2, ds.sizes[lat_dim], ds.sizes[lon_dim])
        time_slice = slice(global_idx, global_idx + 2)
        for var_name in var_names.spatial_resolved_names:
            data = arrays[var_name]
            assert data.shape[0] == 2
            da = ds[var_name]
            if var_name in var_names.time_dependent_names:
                da = da.isel(time=time_slice)
            target_data = as_broadcasted_tensor(da.variable, dims, shape)
            assert torch.equal(data, target_data)

        for var_name in mock_data.var_names.initial_condition_names:
            data = arrays[var_name].detach().numpy()
            assert data.shape[0] == 1
            target_data = ds[var_name][global_idx : global_idx + 1, :, :].values
            assert np.all(data == target_data)


@pytest.mark.parametrize(
    "global_idx",
    [
        pytest.param(31 * 8 - 1, id="monthly_XarrayDataset_2003_03_31_21"),
        pytest.param((31 + 30 + 20) * 8 - 1, id="monthly_XarrayDataset_2003_05_20_21"),
    ],
)
@pytest.mark.parametrize(
    "mock_data_fixture, engine, file_pattern",
    [
        ("mock_monthly_netcdfs", "netcdf4", "*.nc"),
        ("mock_monthly_zarr", "zarr", "*.zarr"),
    ],
)
def test_XarrayDataset_monthly(
    global_idx, mock_data_fixture, engine, file_pattern, request
):
    mock_data: MockData = request.getfixturevalue(mock_data_fixture)
    _test_monthly_values(
        mock_data, global_idx, file_pattern=file_pattern, engine=engine
    )


@pytest.mark.parametrize("n_samples", [None, 1])
def test_XarrayDataset_monthly_n_timesteps(mock_monthly_netcdfs, n_samples):
    """Test that increasing n_timesteps decreases the number of samples."""
    mock_data: MockData = mock_monthly_netcdfs
    if len(mock_data.var_names.initial_condition_names) != 0:
        return
    config = XarrayDataConfig(data_path=mock_data.tmpdir, subset=Slice(stop=n_samples))
    n_forward_steps = 4
    requirements = DataRequirements(
        names=mock_data.var_names.all_names + ["x"],
        n_timesteps=n_forward_steps + 1,
    )
    dataset, _ = get_xarray_dataset(config, requirements)
    if n_samples is None:
        assert len(dataset) == len(mock_data.obs_times) - n_forward_steps
    else:
        assert len(dataset) == n_samples
    assert "x" in dataset[0][0]


@pytest.mark.parametrize(
    "global_idx,expected_file_local_idx",
    [
        pytest.param(365 + 59, (1, 59), id="yearly_file_local_idx_2000_03_01"),
        pytest.param(365, (1, 0), id="yearly_file_local_idx_2000_01_01"),
        pytest.param(364, (0, 364), id="yearly_file_local_idx_1999_12_31"),
    ],
)
def test_yearly_file_local_index(
    mock_yearly_netcdfs, global_idx, expected_file_local_idx
):
    mock_data: MockData = mock_yearly_netcdfs
    file_local_idx = get_file_local_index(global_idx, mock_data.start_indices)
    assert file_local_idx == expected_file_local_idx
    delta = mock_data.obs_times[1] - mock_data.obs_times[0]
    target_timestamp = (
        cftime.DatetimeNoLeap(1999, 1, 1, 0, 0, 0, 0, has_year_zero=True)
        + global_idx * delta
    )
    file_idx, local_idx = file_local_idx
    full_paths = sorted(list(mock_data.tmpdir.glob("*.nc")))
    with xr.open_dataset(full_paths[file_idx], use_cftime=True) as ds:
        assert ds["time"][local_idx].item() == target_timestamp


@pytest.mark.parametrize(
    "global_idx",
    [
        pytest.param(364, id="yearly_XarrayDataset_1999_12_31"),
        pytest.param(365 + 31 + 28, id="yearly_XarrayDataset_2000_02_28"),
    ],
)
def test_XarrayDataset_yearly(mock_yearly_netcdfs, global_idx):
    mock_data: MockData = mock_yearly_netcdfs
    config = XarrayDataConfig(data_path=mock_data.tmpdir)
    with xr.open_mfdataset(
        mock_data.tmpdir.glob("*.nc"),
        use_cftime=True,
        data_vars="minimal",
        coords="minimal",
    ) as ds:
        for n_steps in [3, 50]:
            requirements = DataRequirements(
                names=mock_data.var_names.all_names,
                n_timesteps=n_steps,
            )
            dataset = XarrayDataset(config=config, requirements=requirements)
            assert len(dataset) == len(mock_data.obs_times) - n_steps + 1
            lon_dim, lat_dim = infer_horizontal_dimension_names(ds)
            dims = ("time", lat_dim, lon_dim)
            shape = (n_steps, ds.sizes[lat_dim], ds.sizes[lon_dim])
            time_slice = slice(global_idx, global_idx + n_steps)
            for var_name in mock_data.var_names.spatial_resolved_names:
                da = ds[var_name]
                if var_name in mock_data.var_names.time_dependent_names:
                    da = da.isel(time=time_slice)
                target_data = as_broadcasted_tensor(da.variable, dims, shape)
                target_times = ds["time"][global_idx : global_idx + n_steps].drop_vars(
                    "time"
                )
                data, time = dataset[global_idx]
                data_tensor = data[var_name]
                assert data_tensor.shape[0] == n_steps
                assert torch.equal(data_tensor, target_data)
                xr.testing.assert_equal(time, target_times)


def test_dataset_dtype_casting(mock_monthly_netcdfs):
    mock_data: MockData = mock_monthly_netcdfs
    config = XarrayDataConfig(data_path=mock_data.tmpdir, dtype="bfloat16")
    requirements = DataRequirements(names=mock_data.var_names.all_names, n_timesteps=2)
    dataset = XarrayDataset(config=config, requirements=requirements)
    assert isinstance(dataset.horizontal_coordinates, LatLonCoordinates)
    assert dataset.horizontal_coordinates.lat.dtype == torch.bfloat16
    assert dataset.horizontal_coordinates.lon.dtype == torch.bfloat16
    assert dataset.vertical_coordinate.ak.dtype == torch.bfloat16
    assert dataset.vertical_coordinate.bk.dtype == torch.bfloat16
    data, _ = dataset[0]
    for tensor in data.values():
        assert tensor.dtype == torch.bfloat16


def test_time_invariant_variable_is_repeated(mock_monthly_netcdfs):
    mock_data: MockData = mock_monthly_netcdfs
    config = XarrayDataConfig(data_path=mock_data.tmpdir)
    requirements = DataRequirements(names=mock_data.var_names.all_names, n_timesteps=15)
    dataset = XarrayDataset(config=config, requirements=requirements)
    data = dataset[0][0]
    assert data["constant_var"].shape[0] == 15
    assert data["constant_scalar_var"].shape == (15, 4, 8)


def _get_repeat_dataset(
    mock_data: MockData, n_timesteps: int, n_repeats: int
) -> XarrayDataset:
    config = XarrayDataConfig(data_path=mock_data.tmpdir, n_repeats=n_repeats)
    requirements = DataRequirements(
        names=mock_data.var_names.all_names, n_timesteps=n_timesteps
    )
    return XarrayDataset(config, requirements)


@pytest.mark.parametrize("n_timesteps", [1, 4])
@pytest.mark.parametrize("n_repeats", [1, 2])
def test_repeat_dataset_num_timesteps(
    mock_monthly_netcdfs: MockData, n_timesteps, n_repeats
):
    unrepeated_dataset = _get_repeat_dataset(mock_monthly_netcdfs, n_timesteps, 1)
    data = _get_repeat_dataset(mock_monthly_netcdfs, n_timesteps, n_repeats)
    offset = n_timesteps - 1
    expected_length = n_repeats * (len(unrepeated_dataset) + offset) - offset
    assert len(data) == expected_length


@pytest.mark.parametrize(
    "glob_pattern, expected_num_files, expected_year_month_tuples",
    [
        ("*.nc", None, None),
        ("2003030100.nc", 1, [(2003, 3)]),
        ("2003??0100.nc", 3, [(2003, i) for i in range(3, 6)]),
    ],
    ids=["all_files", "single_file", "all_2003_files"],
)
def test_glob_file_pattern(
    mock_monthly_netcdfs: MockData,
    glob_pattern,
    expected_num_files,
    expected_year_month_tuples,
):
    config = XarrayDataConfig(
        data_path=mock_monthly_netcdfs.tmpdir, file_pattern=glob_pattern
    )
    requirements = DataRequirements(
        names=mock_monthly_netcdfs.var_names.all_names, n_timesteps=2
    )
    dataset = XarrayDataset(config, requirements)
    if expected_num_files is None:
        expected_num_files = len(mock_monthly_netcdfs.start_times)
    assert expected_num_files == len(dataset.full_paths)

    if expected_year_month_tuples is not None:
        for i, (year, month) in enumerate(expected_year_month_tuples):
            assert f"{year}{month:02d}" in dataset.full_paths[i]


def test_time_slice():
    time_slice = TimeSlice("2001-01-01", "2001-01-05", 2)
    time_index = xr.cftime_range("2000", "2002", freq="D", calendar="noleap")
    slice_ = time_slice.slice(time_index)
    assert slice_ == slice(365, 370, 2)


def test_time_index(mock_monthly_netcdfs):
    config = XarrayDataConfig(data_path=mock_monthly_netcdfs.tmpdir)
    n_timesteps = 2
    dataset = XarrayDataset(
        config,
        DataRequirements(
            names=mock_monthly_netcdfs.var_names.all_names, n_timesteps=n_timesteps
        ),
    )
    last_sample_init_time = len(mock_monthly_netcdfs.obs_times) - n_timesteps + 1
    obs_times = mock_monthly_netcdfs.obs_times[:last_sample_init_time]
    assert dataset.sample_start_time.equals(xr.CFTimeIndex(obs_times))


@pytest.mark.parametrize("infer_timestep", [True, False])
def test_XarrayDataset_timestep(mock_monthly_netcdfs, infer_timestep):
    config = XarrayDataConfig(
        data_path=mock_monthly_netcdfs.tmpdir, infer_timestep=infer_timestep
    )
    n_timesteps = 2
    dataset = XarrayDataset(
        config,
        DataRequirements(
            names=mock_monthly_netcdfs.var_names.all_names, n_timesteps=n_timesteps
        ),
    )
    if infer_timestep:
        expected_timestep = pd.Timedelta(MOCK_DATA_FREQ).to_pytimedelta()
        assert dataset.timestep == expected_timestep
    else:
        with pytest.raises(ValueError, match="Timestep was not inferred"):
            assert dataset.timestep


@pytest.mark.parametrize(
    ("periods", "freq", "reverse", "expected", "exception"),
    [
        pytest.param(
            2,
            "3h",
            False,
            datetime.timedelta(hours=3),
            None,
            id="2 timesteps, regular freq",
        ),
        pytest.param(
            3,
            "9h",
            False,
            datetime.timedelta(hours=9),
            None,
            id="3 timesteps, regular freq",
        ),
        pytest.param(3, "3h", True, None, ValueError, id="3 timesteps, negative freq"),
        pytest.param(
            3, "MS", False, None, ValueError, id="3 timesteps, irregular freq"
        ),
        pytest.param(1, "D", False, None, ValueError, id="1 timestep"),
    ],
)
def test_get_timestep(periods, freq, reverse, expected, exception):
    index = xr.cftime_range("2000", periods=periods, freq=freq)

    if reverse:
        index = index[::-1]

    if exception is None:
        result = get_timestep(index.values)
        assert result == expected
    else:
        with pytest.raises(exception):
            get_timestep(index.values)


@pytest.mark.parametrize("n_repeats", [1, 3])
def test_repeat_and_increment_times(n_repeats):
    freq = "5h"
    delta = pd.Timedelta(freq).to_pytimedelta()

    start_a = cftime.DatetimeGregorian(2000, 1, 1)
    periods_a = 2
    segment_a = xr.cftime_range(start_a, periods=periods_a, freq=freq).values

    start_b = segment_a[-1] + delta
    periods_b = 3
    segment_b = xr.cftime_range(start_b, periods=periods_b, freq=freq).values

    raw_times = [segment_a, segment_b]
    raw_periods = [periods_a, periods_b]
    raw_total_periods = sum(raw_periods)

    result = repeat_and_increment_time(raw_times, n_repeats, delta)
    full_periods = [len(times) for times in result]
    full_total_periods = sum(full_periods)

    result_concatenated = np.concatenate(result)
    expected_concatenated = xr.cftime_range(
        start_a, periods=full_total_periods, freq=freq
    ).values

    assert full_periods == n_repeats * raw_periods
    assert full_total_periods == n_repeats * raw_total_periods
    np.testing.assert_equal(result_concatenated, expected_concatenated)


@pytest.mark.parametrize("n_repeats", [1, 3])
def test_all_times(mock_monthly_netcdfs, n_repeats):
    n_timesteps = 2  # Arbitrary for this test
    dataset = _get_repeat_dataset(mock_monthly_netcdfs, n_timesteps, n_repeats)
    expected_periods = n_repeats * len(mock_monthly_netcdfs.obs_times)
    expected = xr.cftime_range(
        MOCK_DATA_START_DATE, periods=expected_periods, freq=MOCK_DATA_FREQ
    )
    result = dataset.all_times
    assert result.equals(expected)


def test_get_sample_by_time_slice_times_n_repeats(mock_monthly_netcdfs: MockData):
    n_timesteps = 2  # Arbitrary for this test
    n_repeats = 3
    repeated_dataset = _get_repeat_dataset(mock_monthly_netcdfs, n_timesteps, n_repeats)

    # Pick a slice that is outside the range of the unrepeated data
    unrepeated_length = len(repeated_dataset.all_times) // n_repeats
    time_slice = slice(unrepeated_length, unrepeated_length + 3)

    _, result = repeated_dataset.get_sample_by_time_slice(time_slice)
    expected = xr.DataArray(
        repeated_dataset.all_times[time_slice].values, dims=["time"]
    )
    xr.testing.assert_equal(result, expected)


@pytest.mark.parametrize(
    "dtype,expected_torch_dtype", [("int16", torch.int16), (None, None)]
)
def test_dataset_config_dtype(dtype, expected_torch_dtype):
    config = XarrayDataConfig(data_path="path/to/data", dtype=dtype)
    assert config.torch_dtype == expected_torch_dtype


def test_dataset_config_dtype_raises():
    with pytest.raises(ValueError):
        XarrayDataConfig(data_path="path/to/data", dtype="invalid_dtype")


def test_renaming(mock_monthly_netcdfs):
    # stepper in/out names should be variables after renaming
    stepper_variables = ["foo", "bar_new"]
    config = XarrayDataConfig(
        data_path=mock_monthly_netcdfs.tmpdir, renamed_variables={"bar": "bar_new"}
    )
    dataset = XarrayDataset(
        config,
        DataRequirements(names=stepper_variables, n_timesteps=2),
    )
    data, _ = dataset[0]
    assert "bar_new" in data
    assert "bar" not in data


def test_fill_nans(mock_monthly_netcdfs_with_nans):
    nan_config = FillNaNsConfig()
    config = XarrayDataConfig(
        data_path=mock_monthly_netcdfs_with_nans.tmpdir, fill_nans=nan_config
    )
    requirements = DataRequirements(
        names=mock_monthly_netcdfs_with_nans.var_names.all_names, n_timesteps=2
    )
    dataset = XarrayDataset(config, requirements)
    data, _ = dataset[0]
    assert torch.all(data["foo"][0, :, 0] == 0)


def test_keep_nans(mock_monthly_netcdfs_with_nans):
    config_keep_nan = XarrayDataConfig(data_path=mock_monthly_netcdfs_with_nans.tmpdir)
    requirements = DataRequirements(
        names=mock_monthly_netcdfs_with_nans.var_names.all_names, n_timesteps=2
    )
    dataset = XarrayDataset(config_keep_nan, requirements)
    data_with_nan, _ = dataset[0]
    assert torch.all(torch.isnan(data_with_nan["foo"][0, :, 0]))


def test_overwrite(mock_monthly_netcdfs):
    const = -10
    multiple = 3.5

    overwrite_config = OverwriteConfig(
        constant={"foo": const},
        multiply_scalar={"bar": multiple},
    )

    config = XarrayDataConfig(data_path=mock_monthly_netcdfs.tmpdir)
    n_timesteps = 2
    dataset = XarrayDataset(
        config,
        DataRequirements(
            names=mock_monthly_netcdfs.var_names.all_names, n_timesteps=n_timesteps
        ),
    )[0][0]

    config_overwrite = XarrayDataConfig(
        data_path=mock_monthly_netcdfs.tmpdir, overwrite=overwrite_config
    )
    n_timesteps = 2
    dataset_overwrite = XarrayDataset(
        config_overwrite,
        DataRequirements(
            names=mock_monthly_netcdfs.var_names.all_names, n_timesteps=n_timesteps
        ),
    )[0][0]

    for v in ["foo", "bar"]:
        assert dataset_overwrite[v].dtype == dataset[v].dtype
        assert dataset_overwrite[v].device == dataset[v].device
    assert torch.equal(
        dataset_overwrite["foo"], torch.ones_like(dataset["foo"]) * const
    )
    assert torch.equal(dataset_overwrite["bar"], dataset["bar"] * multiple)


def test_overwrite_raises_error_on_original_name(mock_monthly_netcdfs):
    overwrite_config = OverwriteConfig(
        constant={"foo": 3},
    )
    with pytest.raises(ValueError):
        XarrayDataConfig(
            data_path=mock_monthly_netcdfs.tmpdir,
            overwrite=overwrite_config,
            renamed_variables={"foo": "foo_new"},
        )



import datetime
import unittest.mock
from collections import namedtuple
from typing import Callable, Iterable, Optional, Tuple

import numpy as np
import pytest
import torch
import xarray as xr

import fme
from fme.ace.data_loading.batch_data import BatchData, PrognosticState
from fme.ace.stepper import SingleModuleStepperConfig
from fme.core.coordinates import HybridSigmaPressureCoordinate
from fme.core.device import get_device
from fme.core.generics.data import SimpleInferenceData
from fme.core.generics.inference import (
    Looper,
    PredictFunction,
    get_record_to_wandb,
    run_inference,
)
from fme.core.gridded_ops import LatLonOperations
from fme.core.loss import WeightedMappingLossConfig
from fme.core.normalizer import NormalizationConfig
from fme.core.registry.module import ModuleSelector
from fme.core.testing.wandb import mock_wandb
from fme.core.timing import GlobalTimer
from fme.core.typing_ import TensorDict, TensorMapping

SphericalData = namedtuple("SphericalData", ["data", "area_weights", "vertical_coord"])


def get_data(
    names: Iterable[str], shape: Tuple[int, int, int, int, int]
) -> SphericalData:
    data = {}
    n_lat = shape[2]
    n_lon = shape[3]
    shape_without_z = shape[:-1]
    nz = shape[-1]
    lats = torch.linspace(-89.5, 89.5, n_lat)
    for name in names:
        data[name] = torch.rand(*shape_without_z, device=fme.get_device())
    area_weights = fme.spherical_area_weights(lats, n_lon).to(fme.get_device())
    ak, bk = torch.arange(nz), torch.arange(nz)
    vertical_coord = HybridSigmaPressureCoordinate(ak, bk)
    return SphericalData(data, area_weights, vertical_coord)


def get_scalar_data(names, value):
    return {n: np.array([value], dtype=np.float32) for n in names}


class MockLoader(torch.utils.data.DataLoader):
    def __init__(
        self,
        shape: tuple,
        names: Iterable[str],
        n_windows: int,
        time: Optional[xr.DataArray] = None,
    ):
        device = fme.get_device()
        self._data = {n: torch.rand(*shape, device=device) for n in names}
        if time is None:
            self._time = xr.DataArray(np.zeros(shape[:2]), dims=["sample", "time"])
        elif time.shape != shape[:2]:
            raise ValueError(
                "Time shape must match the first two dimensions of the data."
            )
        else:
            self._time = time
        self._n_windows = n_windows
        self._current_window = 0

    def __iter__(self):
        return self

    def __len__(self) -> int:
        return self._n_windows

    def __next__(self) -> BatchData:
        if self._current_window < self._n_windows:
            self._current_window += 1
            return BatchData.new_on_device(
                data=self._data,
                time=self._time
                + (self._current_window - 1) * (self._time.shape[1] - 1),
            )
        else:
            raise StopIteration


def _get_stepper():
    class ChannelSum(torch.nn.Module):
        def forward(self, x):
            summed = torch.sum(x, dim=-3, keepdim=True)
            diagnostic = torch.rand_like(summed)
            return torch.concat([x, diagnostic], dim=-3)

    n_samples = 2
    n_time = 5
    n_lat = 2
    n_lon = 4
    nz = 3
    shape = (n_samples, n_time, n_lat, n_lon, nz)
    in_names = ["forcing", "prognostic"]
    out_names = ["prognostic", "diagnostic"]
    all_names = list(set(in_names + out_names))

    spherical_data = get_data(all_names, shape)
    time = xr.DataArray(
        np.repeat(np.expand_dims(np.arange(n_time), axis=0), n_samples, axis=0),
        dims=["sample", "time"],
    )

    img_shape = spherical_data.data[in_names[0]].shape[2:]
    gridded_operations = LatLonOperations(spherical_data.area_weights)
    vertical_coordinate = spherical_data.vertical_coord
    config = SingleModuleStepperConfig(
        builder=ModuleSelector(type="prebuilt", config={"module": ChannelSum()}),
        in_names=in_names,
        out_names=out_names,
        normalization=NormalizationConfig(
            means=get_scalar_data(all_names, 0.0),
            stds=get_scalar_data(all_names, 1.0),
        ),
        loss=WeightedMappingLossConfig(),
    )
    stepper = config.get_stepper(
        img_shape,
        gridded_operations,
        vertical_coordinate,
        datetime.timedelta(seconds=1),
    )
    return stepper, spherical_data, time, in_names, out_names


def test_looper():
    stepper, spherical_data, time, in_names, out_names = _get_stepper()
    forcing_names = set(in_names) - set(out_names)
    shape = spherical_data.data[in_names[0]].shape
    initial_condition = BatchData.new_on_device(
        data={n: spherical_data.data[n][:, :1] for n in spherical_data.data},
        time=time[:, 0:1],
    ).get_start(
        prognostic_names=stepper.prognostic_names,
        n_ic_timesteps=1,
    )
    loader = MockLoader(shape, forcing_names, 3, time=time)
    looper = Looper(
        predict=stepper.predict,
        data=SimpleInferenceData(initial_condition, loader),
    )

    expected_output_shape = (shape[0], shape[1] - 1, shape[2], shape[3])
    for batch in looper:
        assert set(out_names) == set(batch.data)
        for name in out_names:
            assert batch.data[name].shape == expected_output_shape


def test_looper_paired():
    stepper, spherical_data, time, in_names, out_names = _get_stepper()
    forcing_names = set(in_names) - set(out_names)
    shape = spherical_data.data[in_names[0]].shape
    initial_condition = BatchData.new_on_device(
        data={n: spherical_data.data[n][:, :1] for n in spherical_data.data},
        time=time[:, 0:1],
    ).get_start(
        prognostic_names=stepper.prognostic_names,
        n_ic_timesteps=1,
    )
    loader = MockLoader(shape, forcing_names, 3, time=time)
    looper = Looper(
        predict=stepper.predict_paired,
        data=SimpleInferenceData(initial_condition, loader),
    )

    expected_output_shape = (shape[0], shape[1] - 1, shape[2], shape[3])
    for batch in looper:
        assert set(out_names) == set(batch.prediction)
        assert set(forcing_names) == set(batch.target)
        for name in out_names:
            assert batch.prediction[name].shape == expected_output_shape
        for name in forcing_names:
            assert batch.target[name].shape == expected_output_shape


def _mock_compute_derived_quantities(data, forcing_data):
    data_name = list(data)[0]
    forcing_name = list(forcing_data)[0]
    derived = {"derived": data[data_name] + forcing_data[forcing_name]}
    return {**data, **derived}


def test_looper_paired_with_derived_variables():
    mock_derive_func = unittest.mock.MagicMock(
        side_effect=_mock_compute_derived_quantities
    )
    stepper, spherical_data, time, in_names, out_names = _get_stepper()
    stepper.derive_func = mock_derive_func
    forcing_names = set(in_names) - set(out_names)
    shape = spherical_data.data[in_names[0]].shape
    initial_condition = BatchData.new_on_device(
        {n: spherical_data.data[n][:, :1] for n in spherical_data.data},
        time=time[:, 0:1],
    ).get_start(
        prognostic_names=stepper.prognostic_names,
        n_ic_timesteps=1,
    )
    loader = MockLoader(shape, forcing_names, 2, time=time)
    looper = Looper(
        predict=stepper.predict_paired,
        data=SimpleInferenceData(initial_condition, loader),
    )

    for batch in looper:
        assert "derived" in batch.prediction
        assert "derived" in batch.target
    mock_derive_func.assert_called()


def test_looper_paired_with_target_data():
    stepper, spherical_data, time, in_names, out_names = _get_stepper()
    all_names = list(set(in_names + out_names))
    shape = spherical_data.data[in_names[0]].shape
    initial_condition = BatchData.new_on_device(
        data={n: spherical_data.data[n][:, :1] for n in spherical_data.data},
        time=time[:, 0:1],
    ).get_start(
        prognostic_names=stepper.prognostic_names,
        n_ic_timesteps=1,
    )
    loader = MockLoader(shape, all_names, 2, time=time)
    looper = Looper(
        predict=stepper.predict_paired,
        data=SimpleInferenceData(initial_condition, loader),
    )

    for batch in looper:
        assert set(out_names) == set(batch.prediction)
        assert set(all_names) == set(batch.target)


def test_looper_paired_with_target_data_and_derived_variables():
    mock_derive_func = unittest.mock.MagicMock(
        side_effect=_mock_compute_derived_quantities
    )
    stepper, spherical_data, time, in_names, out_names = _get_stepper()
    stepper.derive_func = mock_derive_func
    all_names = list(set(in_names + out_names))
    shape = spherical_data.data[in_names[0]].shape
    initial_condition = BatchData.new_on_device(
        data={n: spherical_data.data[n][:, :1] for n in spherical_data.data},
        time=time[:, 0:1],
    ).get_start(
        prognostic_names=stepper.prognostic_names,
        n_ic_timesteps=1,
    )
    loader = MockLoader(shape, all_names, 2, time=time)
    looper = Looper(
        predict=stepper.predict_paired,
        data=SimpleInferenceData(initial_condition, loader),
    )

    for batch in looper:
        assert set(out_names + ["derived"]) == set(batch.prediction)
        assert set(all_names + ["derived"]) == set(batch.target)
    mock_derive_func.assert_called()


def get_batch_data(
    start_time,
    n_timesteps,
):
    n_samples = 1
    n_lat = 3
    n_lon = 4
    time_values = torch.arange(
        start_time, start_time + n_timesteps, device=get_device()
    )[None, :, None, None]
    time_axis = torch.broadcast_to(
        start_time + torch.arange(n_timesteps)[None, :], (n_samples, n_timesteps)
    )
    time = xr.DataArray(time_axis, dims=["sample", "time"])
    return BatchData.new_on_device(
        data={
            "var": torch.broadcast_to(
                time_values, (n_samples, n_timesteps, n_lat, n_lon)
            )
        },
        time=time,
    )


class PlusOneStepper:
    def __init__(
        self,
        n_ic_timesteps: int,
        derive_func: Optional[
            Callable[[TensorMapping, TensorMapping], TensorDict]
        ] = None,
    ):
        self.n_ic_timesteps = n_ic_timesteps
        if derive_func is None:
            self.derive_func: Callable[[TensorMapping, TensorMapping], TensorDict] = (
                unittest.mock.MagicMock(side_effect=lambda x, y=None: dict(x))
            )
        else:
            self.derive_func = derive_func
        _: PredictFunction[  # for type checking
            PrognosticState,
            BatchData,
            BatchData,
        ] = self.predict

    def predict(
        self,
        initial_condition: PrognosticState,
        forcing: BatchData,
        compute_derived_variables: bool = False,
    ) -> Tuple[BatchData, PrognosticState]:
        ic_state = initial_condition.as_batch_data()
        n_forward_steps = forcing.time.shape[1] - self.n_ic_timesteps
        out_tensor = torch.zeros(
            ic_state.data["var"].shape[0],
            n_forward_steps,
            *ic_state.data["var"].shape[2:],
            device=ic_state.data["var"].device,
            dtype=ic_state.data["var"].dtype,
        )
        out_tensor[:, 0, ...] = ic_state.data["var"][:, -1, ...] + 1
        for i in range(1, n_forward_steps):
            out_tensor[:, i, ...] = out_tensor[:, i - 1, ...] + 1
        data = BatchData.new_on_device(
            data={"var": out_tensor},
            time=forcing.time[:, self.n_ic_timesteps :],
        )
        if compute_derived_variables:
            data = data.compute_derived_variables(
                derive_func=self.derive_func, forcing_data=data
            )
        return data, data.get_end(["var"], self.n_ic_timesteps)

    def get_forward_data(
        self,
        forcing: BatchData,
        compute_derived_variables: bool = False,
    ) -> BatchData:
        if compute_derived_variables:
            forcing = forcing.compute_derived_variables(
                derive_func=self.derive_func, forcing_data=forcing
            )
        return forcing.remove_initial_condition(self.n_ic_timesteps)


def test_looper_simple_batch_data():
    n_ic_timesteps = 1
    n_forward_steps = 2
    n_iterations = 10
    mock_derive_func = unittest.mock.MagicMock(
        side_effect=lambda batch_data, forcing_data: batch_data
    )
    stepper = PlusOneStepper(
        n_ic_timesteps=n_ic_timesteps, derive_func=mock_derive_func
    )
    initial_condition = get_batch_data(
        0,
        n_timesteps=n_ic_timesteps,
    ).get_start(prognostic_names=["var"], n_ic_timesteps=n_ic_timesteps)
    loader = [
        get_batch_data(
            i,
            n_ic_timesteps + n_forward_steps,
        )
        for i in range(0, n_iterations * n_forward_steps, n_forward_steps)
    ]

    mock_predict = unittest.mock.MagicMock(side_effect=stepper.predict)
    with unittest.mock.patch.object(stepper, "predict", mock_predict):
        with GlobalTimer():
            timer = GlobalTimer.get_instance()
            looper = Looper(
                predict=stepper.predict,
                data=SimpleInferenceData(initial_condition, loader),
            )
            for i, batch in enumerate(looper):
                for j in range(batch.time.shape[1]):
                    assert torch.allclose(
                        batch.data["var"][:, j, ...],
                        torch.as_tensor(n_ic_timesteps + i * n_forward_steps + j),
                    )
            times = timer.get_durations()
            assert times["data_loading"] > 0
            assert mock_derive_func.call_count == n_iterations
            assert mock_predict.call_count == n_iterations
            # we mocked out the implicit calls that happen in .predict
            # if this changed, update the test
            assert "forward_prediction" not in times
            assert "compute_derived_variables" not in times


def get_mock_aggregator(
    n_ic_timesteps: int,
) -> unittest.mock.MagicMock:
    mock_aggregator = unittest.mock.MagicMock()

    # record_batch will start at step n_ic_timesteps
    i = n_ic_timesteps

    def record_batch_side_effect(
        data: BatchData,
    ):
        nonlocal i
        ret = [{"step": j} for j in range(i, i + data.time.shape[1])]
        i += data.time.shape[1]
        return ret

    mock_aggregator = unittest.mock.MagicMock()
    mock_aggregator.record_initial_condition = unittest.mock.MagicMock(
        return_value=[{"step": j} for j in range(n_ic_timesteps)]
    )

    def get_summary_logs_side_effect():
        # we expect this gets called _outside_ of run_inference
        raise ValueError("should not be called")

    mock_aggregator.get_summary_logs = unittest.mock.MagicMock(
        side_effect=get_summary_logs_side_effect
    )
    mock_aggregator.record_batch = unittest.mock.MagicMock(
        side_effect=record_batch_side_effect
    )
    return mock_aggregator


def get_mock_writer() -> unittest.mock.MagicMock:
    mock_writer = unittest.mock.MagicMock()
    return mock_writer


@pytest.mark.parametrize(
    "n_ic_timesteps, n_forward_steps, n_iterations",
    [
        pytest.param(1, 2, 5, id="n_ic_timesteps=1"),
        pytest.param(2, 2, 5, id="n_ic_timesteps=2"),
    ],
)
def test_run_inference_simple(
    n_ic_timesteps: int, n_forward_steps: int, n_iterations: int
):
    mock_derive_func = unittest.mock.MagicMock(
        side_effect=lambda batch_data, forcing_data: batch_data
    )
    stepper = PlusOneStepper(
        n_ic_timesteps=n_ic_timesteps, derive_func=mock_derive_func
    )
    initial_condition = get_batch_data(
        0,
        n_timesteps=n_ic_timesteps,
    ).get_start(prognostic_names=["var"], n_ic_timesteps=n_ic_timesteps)
    loader = [
        get_batch_data(
            i,
            n_ic_timesteps + n_forward_steps,
        )
        for i in range(0, n_iterations * n_forward_steps, n_forward_steps)
    ]
    mock_writer = get_mock_writer()
    mock_aggregator = get_mock_aggregator(n_ic_timesteps)

    with GlobalTimer():
        with mock_wandb() as wandb:
            wandb.configure(log_to_wandb=True)
            record_logs = unittest.mock.MagicMock(
                side_effect=get_record_to_wandb("inference")
            )  # this init must be within mock_wandb context
            run_inference(
                predict=stepper.predict,
                data=SimpleInferenceData(initial_condition, loader),
                writer=mock_writer,
                aggregator=mock_aggregator,
                record_logs=record_logs,
            )
            wandb_logs = wandb.get_logs()
        timer = GlobalTimer.get_instance()
        times = timer.get_durations()
        assert times["wandb_logging"] > 0
        assert times["data_writer"] > 0
        assert times["aggregator"] > 0
        assert mock_writer.write.call_count == 2
        assert mock_aggregator.record_initial_condition.call_count == 1
        assert mock_writer.append_batch.call_count == n_iterations
        assert mock_aggregator.record_batch.call_count == n_iterations
        assert len(wandb_logs) == n_ic_timesteps + n_iterations * n_forward_steps
        assert wandb_logs == [
            {"inference/step": i}
            for i in range(n_ic_timesteps + n_iterations * n_forward_steps)
        ]
        assert (
            record_logs.call_count == n_iterations + 1
        )  # +1 for the initial condition



from typing import Any, Dict, Type

import pytest
import torch

from fme.core.gridded_ops import GriddedOperations, HEALPixOperations, LatLonOperations


@pytest.mark.parametrize(
    "state, expected_class",
    [
        (
            {
                "type": "LatLonOperations",
                "state": {"area_weights": torch.tensor([1.0, 2.0])},
            },
            LatLonOperations,
        ),
        (
            {
                "type": "HEALPixOperations",
                "state": {},
            },
            HEALPixOperations,
        ),
    ],
)
def test_gridded_operations_from_state(
    state: Dict[str, Any],
    expected_class: Type[GriddedOperations],
):
    ops = GriddedOperations.from_state(state)
    assert isinstance(ops, expected_class)

    recovered_state = ops.to_state()
    assert recovered_state == state

    with pytest.raises(RuntimeError):
        expected_class.from_state(state["state"])



import abc
from collections import namedtuple
from typing import Tuple

import torch
import xarray as xr

from fme.core.typing_ import TensorDict

VariableMetadata = namedtuple("VariableMetadata", ["units", "long_name"])


class Dataset(torch.utils.data.Dataset, abc.ABC):
    @abc.abstractmethod
    def get_sample_by_time_slice(
        self, time_slice: slice
    ) -> Tuple[TensorDict, xr.DataArray]:
        """
        Returns a sample of data for the given time slice.

        Args:
            time_slice: The time slice to return data for.

        Returns:
            A tuple whose first item is a mapping from variable
            name to tensor of shape [n_time, n_lat, n_lon] and
            whose second item is a time coordinate array.
        """
        ...



"""Integration test for inference entrypoint."""

import dataclasses
import datetime
import pathlib
from typing import List

import cftime
import numpy as np
import pytest
import torch
import xarray as xr
import yaml

import fme
from fme.ace.data_loading.batch_data import PrognosticState
from fme.ace.data_loading.inference import (
    ExplicitIndices,
    ForcingDataLoaderConfig,
    InferenceInitialConditionIndices,
    TimestampList,
)
from fme.ace.inference.data_writer import DataWriterConfig
from fme.ace.inference.inference import (
    InferenceConfig,
    InitialConditionConfig,
    get_initial_condition,
    main,
)
from fme.ace.registry import ModuleSelector
from fme.ace.stepper import SingleModuleStepperConfig
from fme.ace.testing import DimSizes, FV3GFSData
from fme.core.coordinates import (
    DimSize,
    HybridSigmaPressureCoordinate,
    LatLonCoordinates,
)
from fme.core.gridded_ops import LatLonOperations
from fme.core.logging_utils import LoggingConfig
from fme.core.normalizer import NormalizationConfig
from fme.core.testing import mock_wandb

TIMESTEP = datetime.timedelta(hours=6)


class PlusOne(torch.nn.Module):
    def forward(self, x):
        return x + 1


def save_stepper(
    path: pathlib.Path,
    in_names: List[str],
    out_names: List[str],
    mean: float,
    std: float,
    data_shape: List[int],
    timestep: datetime.timedelta = TIMESTEP,
):
    all_names = list(set(in_names).union(out_names))
    config = SingleModuleStepperConfig(
        builder=ModuleSelector(type="prebuilt", config={"module": PlusOne()}),
        in_names=in_names,
        out_names=out_names,
        normalization=NormalizationConfig(
            means={name: mean for name in all_names},
            stds={name: std for name in all_names},
        ),
    )
    area = torch.ones(data_shape[-2:], device=fme.get_device())
    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.arange(7), bk=torch.arange(7)
    )
    stepper = config.get_stepper(
        img_shape=(data_shape[-2], data_shape[-1]),
        gridded_operations=LatLonOperations(area),
        vertical_coordinate=vertical_coordinate,
        timestep=timestep,
    )
    torch.save({"stepper": stepper.get_state()}, path)


def test_inference_entrypoint(tmp_path: pathlib.Path):
    forward_steps_in_memory = 2
    # NOTE: number of inputs and outputs has to be the same for the PlusOne
    # stepper module to work properly
    in_names = ["prog", "forcing_var", "DSWRFtoa"]
    out_names = ["prog", "ULWRFtoa", "USWRFtoa"]
    stepper_path = tmp_path / "stepper"
    horizontal = [DimSize("grid_yt", 16), DimSize("grid_xt", 32)]

    dim_sizes = DimSizes(
        n_time=9,
        horizontal=horizontal,
        nz_interface=4,
    )
    save_stepper(
        stepper_path,
        in_names=in_names,
        out_names=out_names,
        mean=0.0,
        std=1.0,
        data_shape=dim_sizes.shape_nd,
    )
    data = FV3GFSData(
        path=tmp_path,
        names=["forcing_var", "DSWRFtoa"],
        dim_sizes=dim_sizes,
        timestep_days=0.25,
        save_vertical_coordinate=False,
    )
    initial_condition = xr.Dataset(
        {
            "prog": xr.DataArray(
                np.random.rand(2, 16, 32), dims=["sample", "lat", "lon"]
            ),
            "forcing": xr.DataArray(
                np.random.rand(2, 16, 32), dims=["sample", "lat", "lon"]
            ),
            "DSWRFtoa": xr.DataArray(
                np.random.rand(2, 16, 32), dims=["sample", "lat", "lon"]
            ),
        }
    )

    initial_condition_path = tmp_path / "init_data" / "ic.nc"
    initial_condition_path.parent.mkdir()
    initial_condition["time"] = xr.DataArray(
        [
            cftime.DatetimeProlepticGregorian(2000, 1, 1, 6),
            cftime.DatetimeProlepticGregorian(2000, 1, 1, 18),
        ],
        dims=["time"],
    )
    initial_condition.to_netcdf(initial_condition_path, mode="w")
    forcing_loader = ForcingDataLoaderConfig(
        dataset=data.inference_data_loader_config.dataset,
        num_data_workers=data.inference_data_loader_config.num_data_workers,
    )

    config = InferenceConfig(
        experiment_dir=str(tmp_path),
        n_forward_steps=4,
        forward_steps_in_memory=forward_steps_in_memory,
        checkpoint_path=str(stepper_path),
        logging=LoggingConfig(
            log_to_screen=True,
            log_to_file=False,
            log_to_wandb=True,
        ),
        initial_condition=InitialConditionConfig(path=str(initial_condition_path)),
        forcing_loader=forcing_loader,
        data_writer=DataWriterConfig(save_prediction_files=True),
    )
    config_filename = tmp_path / "config.yaml"
    with open(config_filename, "w") as f:
        yaml.dump(dataclasses.asdict(config), f)

    with mock_wandb() as wandb:
        wandb.configure(log_to_wandb=True)
        main(yaml_config=str(config_filename))
        wandb_logs = wandb.get_logs()

    n_ic_timesteps = 1
    summary_log_step = 1
    assert len(wandb_logs) == n_ic_timesteps + config.n_forward_steps + summary_log_step
    for i, log in enumerate(wandb_logs):
        for metric, val in log.items():
            # check that time series metrics match
            if "inference/mean" in metric:
                if i > 0:
                    assert metric in wandb_logs[i]
                    if np.isnan(val):
                        assert np.isnan(wandb_logs[i][metric])
                    else:
                        assert wandb_logs[i][metric] == val
                elif not np.isnan(val):  # for IC only valid data is reported to wandb
                    assert metric in wandb_logs[i]
                    assert wandb_logs[i][metric] == val

    ds = xr.open_dataset(tmp_path / "autoregressive_predictions.nc")
    # prognostic in
    assert "prog" in ds
    # diags in
    assert "ULWRFtoa" in ds
    assert "USWRFtoa" in ds
    # derived in
    assert "net_energy_flux_toa_into_atmosphere" in ds
    # forcings not in
    assert "DSWRFtoa" not in ds
    assert "forcing_var" not in ds
    assert ds["prog"].sizes == {"time": 4, "sample": 2, "lat": 16, "lon": 32}
    np.testing.assert_allclose(
        ds["prog"].isel(time=0).values, initial_condition["prog"].values + 1
    )
    np.testing.assert_allclose(
        ds["prog"].isel(time=1).values, ds["prog"].isel(time=0).values + 1, rtol=1e-6
    )
    saved_data = xr.open_dataset(data.data_filename)
    ops = LatLonCoordinates(
        lat=torch.as_tensor(saved_data["grid_yt"].values),
        lon=torch.as_tensor(saved_data["grid_xt"].values),
    ).gridded_operations
    # check that inference logs match raw output
    for i in range(1, config.n_forward_steps + 1):
        for log_name in wandb_logs[i]:
            if "inference/mean/weighted_mean_gen" in log_name:
                variable_name = log_name.split("/")[-1]
                # note raw output does not include initial condition, hence
                # i-1 below. Code uses area from data, not stepper above.
                raw_variable = ds[variable_name].isel(time=i - 1)
                raw_global_mean = ops.area_weighted_mean(
                    torch.as_tensor(raw_variable.values)
                ).mean()
                np.testing.assert_allclose(
                    raw_global_mean, wandb_logs[i][log_name], rtol=1e-6
                )
    assert "inference/total_steps_per_second" in wandb_logs[-1]


def test_get_initial_condition():
    time_da = xr.DataArray([0, 5], dims=["sample"])
    prognostic_da = xr.DataArray(
        np.random.rand(2, 16, 32), dims=["sample", "lat", "lon"]
    )
    data = xr.Dataset({"prog": prognostic_da, "time": time_da})
    initial_condition = get_initial_condition(data, ["prog"])
    assert isinstance(initial_condition, PrognosticState)
    batch_data = initial_condition.as_batch_data()
    assert batch_data.time.shape == (2, 1)
    initial_times = batch_data.time.isel(time=0)
    assert initial_times.shape == (2,)
    assert initial_times[0] == 0
    assert initial_times[1] == 5
    assert batch_data.data["prog"].shape == (2, 1, 16, 32)
    np.testing.assert_allclose(
        batch_data.data["prog"].squeeze(dim=1).cpu().numpy(), data["prog"].values
    )
    assert batch_data.time.isel(time=0).equals(initial_times)


def test_get_initial_condition_raises_bad_variable_shape():
    time_da = xr.DataArray([0, 5], dims=["sample"])
    prognostic_da = xr.DataArray(
        np.random.rand(2, 8, 16, 32), dims=["sample", "layer", "lat", "lon"]
    )
    data = xr.Dataset({"prog": prognostic_da, "time": time_da})
    with pytest.raises(ValueError):
        get_initial_condition(data, ["prog"])


def test_get_initial_condition_raises_missing_time():
    prognostic_da = xr.DataArray(
        np.random.rand(2, 16, 32), dims=["sample", "lat", "lon"]
    )
    data = xr.Dataset({"prog": prognostic_da})
    with pytest.raises(ValueError):
        get_initial_condition(data, ["prog"])


def test_get_initial_condition_raises_mismatched_time_length():
    time_da = xr.DataArray([0, 5, 10], dims=["time"])
    prognostic_da = xr.DataArray(
        np.random.rand(2, 16, 32), dims=["sample", "lat", "lon"]
    )
    data = xr.Dataset({"prog": prognostic_da, "time": time_da})
    with pytest.raises(ValueError):
        get_initial_condition(data, ["prog"])


@pytest.mark.parametrize(
    "start_indices, expected_time_values",
    [
        (
            None,
            [
                cftime.datetime(2000, 1, 1, 6),
                cftime.datetime(2000, 1, 1, 12),
                cftime.datetime(2000, 1, 1, 18),
            ],
        ),
        (
            InferenceInitialConditionIndices(
                n_initial_conditions=2, first=0, interval=2
            ),
            [cftime.datetime(2000, 1, 1, 6), cftime.datetime(2000, 1, 1, 18)],
        ),
        (
            ExplicitIndices([0, 2]),
            [cftime.datetime(2000, 1, 1, 6), cftime.datetime(2000, 1, 1, 18)],
        ),
        (
            TimestampList(times=["2000-01-01T06:00:00", "2000-01-01T18:00:00"]),
            [cftime.datetime(2000, 1, 1, 6), cftime.datetime(2000, 1, 1, 18)],
        ),
        (
            ExplicitIndices([1]),
            [
                cftime.datetime(2000, 1, 1, 12),
            ],
        ),
    ],
)
def test__subselect_initial_conditions(tmp_path, start_indices, expected_time_values):
    initial_condition = xr.Dataset(
        {"prog": xr.DataArray(np.random.rand(3, 16, 32), dims=["sample", "lat", "lon"])}
    )
    initial_condition_path = tmp_path / "init_data" / "ic.nc"
    initial_condition_path.parent.mkdir()
    initial_condition["time"] = xr.DataArray(
        [
            cftime.datetime(2000, 1, 1, 6),
            cftime.datetime(2000, 1, 1, 12),
            cftime.datetime(2000, 1, 1, 18),
        ],
        dims=["sample"],
    )
    initial_condition.to_netcdf(initial_condition_path, mode="w")

    ic_config = InitialConditionConfig(
        path=str(initial_condition_path),
        start_indices=start_indices,
    )
    ic_data = ic_config.get_dataset()

    np.testing.assert_array_equal(ic_data.time.values, np.array(expected_time_values))
    assert ic_data["prog"].shape == (len(expected_time_values), 16, 32)



import pytest
import torch

from fme import get_device

from .distributed import pad_tensor_at_end, unpad_tensor_at_end


@pytest.mark.parametrize(
    ["padding", "fill_value"],
    [
        pytest.param([0, 0, 0], None, id="no_padding"),
        pytest.param([1, 1, 1], 0.0, id="padding_1"),
        pytest.param([1, 1, 1], 1.0, id="padding_1_fill_one"),
    ],
)
def test_pad_tensor_at_end(padding, fill_value):
    tensor = torch.ones(2, 3, 4)
    padded_tensor = pad_tensor_at_end(tensor, padding, fill_value)
    assert padded_tensor.size() == (2 + padding[0], 3 + padding[1], 4 + padding[2])
    for dim, pad in enumerate(padding):
        if pad > 0:
            assert torch.allclose(
                padded_tensor.select(dim=dim, index=padded_tensor.size(dim) - 1),
                torch.tensor(fill_value),
            )


@pytest.mark.parametrize(
    ["padding"],
    [
        pytest.param([0, 0, 0], id="no_padding"),
        pytest.param([1, 1, 1], id="padding_1"),
    ],
)
def test_pad_unpad_rountrip(padding):
    tensor = torch.ones(2, 3, 4, device=get_device())
    padded_tensor = pad_tensor_at_end(tensor, padding)
    unpadded_tensor = unpad_tensor_at_end(padded_tensor, padding)
    assert unpadded_tensor.size() == tensor.size()
    assert torch.allclose(unpadded_tensor, tensor)



import pytest
import torch

import fme
from fme.core.masking import MaskingConfig
from fme.core.stacker import Stacker


def test_masking_config():
    config = MaskingConfig(
        mask_name="a",
        surface_mask_name="b",
        mask_value=1,
        fill_value=0.0,
    )
    mask = config.build()
    assert mask.mask_name == "a"
    assert mask.mask_value == 1
    assert mask.fill_value == 0.0

    with pytest.raises(ValueError) as err:
        _ = MaskingConfig(
            mask_name="a",
            surface_mask_name="b",
            mask_value=3,
            fill_value=0.0,
        )
    assert "mask_value must be either 0 or 1" in str(err.value)


_SIZE = (4, 4)

_MASK_DATA = {
    "surface_mask": torch.ones(_SIZE, device=fme.get_device()),
    "mask_0": torch.ones(_SIZE, device=fme.get_device()),
    "mask_1": torch.ones(_SIZE, device=fme.get_device()),
}
_MASK_DATA["surface_mask"][1, 1] = 0
_MASK_DATA["mask_0"][0, :] = 0
_MASK_DATA["mask_1"][1, :] = 0


_DATA = {
    "PRESsfc": 10.0 + torch.rand(size=_SIZE, device=fme.get_device()),
    "specific_total_water_0": torch.rand(size=_SIZE, device=fme.get_device()),
    "specific_total_water_1": torch.rand(size=_SIZE, device=fme.get_device()),
}


def test_masking():
    config = MaskingConfig(
        mask_name="mask",
        surface_mask_name="surface_mask",
        mask_value=0,
        fill_value=0.0,
    )
    mask = config.build()
    stacker = Stacker(
        {
            "surface_pressure": ["PRESsfc", "PS"],
            "specific_total_water": ["specific_total_water_"],
        }
    )
    output = mask(stacker, _DATA, _MASK_DATA)
    assert output["PRESsfc"][1, 1] == 0.0
    assert output["PRESsfc"][0, 1] != 0.0
    assert torch.all(output["specific_total_water_0"][0, :] == 0.0)
    assert torch.all(output["specific_total_water_1"][1, :] == 0.0)
    assert torch.all(output["specific_total_water_1"][0, :] != 0.0)


def test_masking_no_3d_masking():
    config = MaskingConfig(
        mask_name="surface_mask",
        mask_value=0,
        fill_value=0.0,
    )
    mask = config.build()
    stacker = Stacker({"surface_pressure": ["PRESsfc", "PS"]})
    output = mask(stacker, _DATA, _MASK_DATA)
    assert output["PRESsfc"][1, 1] == 0.0
    assert output["PRESsfc"][0, 1] != 0.0
    assert torch.all(output["specific_total_water_0"][0, :] != 0.0)
    assert torch.all(output["specific_total_water_1"][1, :] != 0.0)
    assert torch.all(output["specific_total_water_1"][0, :] != 0.0)


def test_masking_no_surface_masking():
    config = MaskingConfig(
        mask_name="mask",
        mask_value=0,
        fill_value=0.0,
    )
    mask = config.build()
    stacker = Stacker({"specific_total_water": ["specific_total_water_"]})
    output = mask(stacker, _DATA, _MASK_DATA)
    assert output["PRESsfc"][1, 1] != 0.0
    assert output["PRESsfc"][0, 1] != 0.0
    assert torch.all(output["specific_total_water_0"][0, :] == 0.0)
    assert torch.all(output["specific_total_water_1"][1, :] == 0.0)
    assert torch.all(output["specific_total_water_1"][0, :] != 0.0)


def test_masking_missing_2d_mask():
    config = MaskingConfig(
        mask_name="mask",
        mask_value=0,
        fill_value=0.0,
    )
    mask = config.build()
    stacker = Stacker(
        {
            "surface_pressure": ["PRESsfc", "PS"],
            "specific_total_water": ["specific_total_water_"],
        }
    )
    with pytest.raises(RuntimeError) as err:
        _ = mask(stacker, _DATA, _MASK_DATA)
    assert "surface_mask_name is None" in str(err.value)
    assert "surface_pressure" in str(err.value)



import pytest
import torch

from fme.core.device import move_tensordict_to_device
from fme.core.normalizer import NormalizationConfig, StandardNormalizer


def test_normalize_depends_on_mean():
    means = {"a": torch.tensor(1.0), "b": torch.tensor(2.0)}
    stds = {"a": torch.tensor(1.0), "b": torch.tensor(1.0)}
    normalizer = StandardNormalizer(means=means, stds=stds)
    tensors = {"a": torch.tensor(1.0), "b": torch.tensor(1.0)}
    normalized = normalizer.normalize(tensors)
    assert normalized["a"] == torch.tensor(0.0)
    assert normalized["b"] == torch.tensor(-1.0)


def test_normalize_depends_on_std():
    means = {"a": torch.tensor(0.0), "b": torch.tensor(0.0)}
    stds = {"a": torch.tensor(1.0), "b": torch.tensor(2.0)}
    normalizer = StandardNormalizer(means=means, stds=stds)
    tensors = {"a": torch.tensor(1.0), "b": torch.tensor(1.0)}
    normalized = normalizer.normalize(tensors)
    assert normalized["a"] == torch.tensor(1.0)
    assert normalized["b"] == torch.tensor(0.5)


def test_denormalize_depends_on_mean():
    means = {"a": torch.tensor(1.0), "b": torch.tensor(2.0)}
    stds = {"a": torch.tensor(1.0), "b": torch.tensor(1.0)}
    normalizer = StandardNormalizer(means=means, stds=stds)
    tensors = {"a": torch.tensor(0.0), "b": torch.tensor(-1.0)}
    denormalized = normalizer.denormalize(tensors)
    assert denormalized["a"] == torch.tensor(1.0)
    assert denormalized["b"] == torch.tensor(1.0)


def test_denormalize_depends_on_std():
    means = {"a": torch.tensor(0.0), "b": torch.tensor(0.0)}
    stds = {"a": torch.tensor(1.0), "b": torch.tensor(2.0)}
    normalizer = StandardNormalizer(means=means, stds=stds)
    tensors = {"a": torch.tensor(1.0), "b": torch.tensor(0.5)}
    denormalized = normalizer.denormalize(tensors)
    assert denormalized["a"] == torch.tensor(1.0)
    assert denormalized["b"] == torch.tensor(1.0)


def test_normalize_and_denormalize_random_tensor():
    torch.manual_seed(0)
    # randomly set means and stds
    means = move_tensordict_to_device({"a": torch.randn(1), "b": torch.randn(1)})
    stds = move_tensordict_to_device({"a": torch.randn(1), "b": torch.randn(1)})
    normalizer = StandardNormalizer(means=means, stds=stds)
    tensors = move_tensordict_to_device({"a": torch.randn(10), "b": torch.randn(10)})
    denormalized = normalizer.denormalize(normalizer.normalize(tensors))
    assert torch.allclose(denormalized["a"], tensors["a"])
    assert torch.allclose(denormalized["b"], tensors["b"])


def test_missing_normalization_build_raises_error():
    normalization = NormalizationConfig(
        means={"a": 1.0, "b": 2.0},
        stds={"a": 1.0, "b": 1.0},
    )
    all_names = ["a", "b", "c"]
    with pytest.raises(KeyError):
        normalization.build(all_names)


def test_tensors_with_missing_normalization_stats_get_filtered():
    normalization = NormalizationConfig(
        means={"a": 1.0, "b": 2.0},
        stds={"a": 1.0, "b": 1.0},
    ).build(["a", "b"])
    sample_input = {"a": torch.zeros(1), "b": torch.zeros(1), "c": torch.zeros(1)}
    sample_input = move_tensordict_to_device(sample_input)

    normalized = normalization.normalize(sample_input)
    assert "c" not in normalized

    denormalized = normalization.denormalize(sample_input)
    assert "c" not in denormalized



import numpy as np
import pytest
import torch
import xarray as xr

from fme.ace.aggregator.one_step import OneStepAggregator
from fme.ace.stepper import TrainOutput
from fme.core.device import get_device
from fme.core.gridded_ops import LatLonOperations


def test_labels_exist():
    n_sample = 10
    n_time = 3
    nx, ny = 2, 2
    loss = 1.0
    area_weights = torch.ones(ny).to(get_device())
    agg = OneStepAggregator(LatLonOperations(area_weights))
    target_data = {"a": torch.randn(n_sample, n_time, nx, ny, device=get_device())}
    gen_data = {"a": torch.randn(n_sample, n_time, nx, ny, device=get_device())}
    agg.record_batch(
        batch=TrainOutput(
            metrics={"loss": loss},
            target_data=target_data,
            gen_data=gen_data,
            time=xr.DataArray(np.zeros((n_sample, n_time)), dims=["sample", "time"]),
            normalize=lambda x: x,
        ),
    )
    logs = agg.get_logs(label="test")
    assert "test/mean/loss" in logs
    assert "test/mean/weighted_rmse/a" in logs
    assert "test/mean/weighted_bias/a" in logs
    assert "test/mean/weighted_grad_mag_percent_diff/a" in logs
    assert "test/snapshot/image-full-field/a" in logs
    assert "test/snapshot/image-residual/a" in logs
    assert "test/snapshot/image-error/a" in logs


def test_aggregator_raises_on_no_data():
    """
    Basic test the aggregator combines loss correctly
    with multiple batches and no distributed training.
    """
    ny = 2
    area_weights = torch.ones(ny).to(get_device())
    agg = OneStepAggregator(LatLonOperations(area_weights))
    with pytest.raises(ValueError) as excinfo:
        agg.record_batch(
            batch=TrainOutput(
                metrics={"loss": 1.0},
                target_data={},
                gen_data={},
                time=xr.DataArray(np.zeros((0, 0)), dims=["sample", "time"]),
                normalize=lambda x: x,
            ),
        )
        # check that the raised exception contains the right substring
        assert "No data" in str(excinfo.value)


def test__get_loss_scaled_mse_components():
    loss_scaling = {
        "a": torch.tensor(1.0),
        "b": torch.tensor(0.5),
    }
    agg = OneStepAggregator(
        gridded_operations=LatLonOperations(
            area_weights=torch.ones(10).to(get_device())
        ),
        loss_scaling=loss_scaling,
    )

    logs = {
        "test/mean/weighted_rmse/a": 1.0,
        "test/mean/weighted_rmse/b": 4.0,
        "test/mean/weighted_rmse/c": 0.0,
    }
    result = agg._get_loss_scaled_mse_components(logs, "test")
    scaled_squared_errors_sum = (1.0 / 1.0) ** 2 + (4.0 / 0.5) ** 2
    assert (
        result["test/mean/mse_fractional_components/a"] == 1 / scaled_squared_errors_sum
    )
    assert (
        result["test/mean/mse_fractional_components/b"]
        == 64 / scaled_squared_errors_sum
    )
    assert "test/mean/mse_fractional_components/c" not in result



import glob
import os
import shutil
import time
from typing import Any, Mapping, Optional

import numpy as np
import wandb

from fme.core.distributed import Distributed

singleton: Optional["WandB"] = None


class DirectInitializationError(RuntimeError):
    pass


class Histogram(wandb.Histogram):
    def __init__(
        self,
        *args,
        direct_access=True,
        **kwargs,
    ):
        if direct_access:
            raise DirectInitializationError(
                "must initialize from `wandb = WandB.get_instance()`, "
                "not directly from `import fme.core.wandb`"
            )
        super().__init__(*args, **kwargs)


Histogram.__doc__ = wandb.Histogram.__doc__
Histogram.__init__.__doc__ = wandb.Histogram.__init__.__doc__


class Table(wandb.Table):
    def __init__(
        self,
        *args,
        direct_access=True,
        **kwargs,
    ):
        if direct_access:
            raise DirectInitializationError(
                "must initialize from `wandb = WandB.get_instance()`, "
                "not directly from `import fme.core.wandb`"
            )
        super().__init__(*args, **kwargs)


Table.__doc__ = wandb.Table.__doc__
Table.__init__.__doc__ = wandb.Table.__init__.__doc__


class Video(wandb.Video):
    def __init__(
        self,
        *args,
        direct_access=True,
        **kwargs,
    ):
        if direct_access:
            raise DirectInitializationError(
                "must initialize from `wandb = WandB.get_instance()`, "
                "not directly from `import fme.core.wandb`"
            )
        super().__init__(*args, **kwargs)


Video.__doc__ = wandb.Video.__doc__
Video.__init__.__doc__ = wandb.Video.__init__.__doc__


class Image(wandb.Image):
    def __init__(
        self,
        *args,
        direct_access=True,
        **kwargs,
    ):
        if direct_access:
            raise DirectInitializationError(
                "must initialize from `wandb = WandB.get_instance()`, "
                "not directly from `import fme.core.wandb`"
            )
        super().__init__(*args, **kwargs)


Image.__doc__ = wandb.Image.__doc__
Image.__init__.__doc__ = wandb.Image.__init__.__doc__


class WandB:
    """
    A singleton class to interface with Weights and Biases (WandB).
    """

    @classmethod
    def get_instance(cls) -> "WandB":
        """
        Get the singleton instance of the WandB class.
        """
        global singleton
        if singleton is None:
            singleton = cls()
        return singleton

    def __init__(self):
        self._enabled = False
        self._configured = False

    def configure(self, log_to_wandb: bool):
        dist = Distributed.get_instance()
        self._enabled = log_to_wandb and dist.is_root()
        self._configured = True

    def init(self, **kwargs):
        """Kwargs are passed to wandb.init."""
        if not self._configured:
            raise RuntimeError(
                "must call WandB.configure before WandB init can be called"
            )
        if self._enabled:
            wandb.require("core")
            wandb.init(**kwargs)

    def watch(self, modules):
        if self._enabled:
            wandb.watch(modules)

    def log(self, data: Mapping[str, Any], step=None, sleep=None):
        if self._enabled:
            wandb.log(data, step=step)
            if sleep is not None:
                time.sleep(sleep)

    def Image(self, data_or_path, *args, **kwargs) -> Image:
        if isinstance(data_or_path, np.ndarray):
            data_or_path = scale_image(data_or_path)

        return Image(data_or_path, *args, direct_access=False, **kwargs)

    def clean_wandb_dir(self, experiment_dir: str):
        # this is necessary because wandb does not remove run media directories
        # after a run is synced; see https://github.com/wandb/wandb/issues/3564
        if self._enabled:
            wandb.run.finish()  # necessary to ensure the run directory is synced
            wandb_dir = os.path.join(experiment_dir, "wandb")
            remove_media_dirs(wandb_dir)

    def Video(self, *args, **kwargs) -> Video:
        return Video(*args, direct_access=False, **kwargs)

    def Table(self, *args, **kwargs) -> Table:
        return Table(*args, direct_access=False, **kwargs)

    def Histogram(self, *args, **kwargs) -> Histogram:
        return Histogram(*args, direct_access=False, **kwargs)

    @property
    def enabled(self) -> bool:
        return self._enabled


def scale_image(
    image_data: np.ndarray,
) -> np.ndarray:
    """
    Given an array of scalar data, rescale the data to the range [0, 255].
    """
    data_min = np.nanmin(image_data)
    data_max = np.nanmax(image_data)
    # video data is brightness values on a 0-255 scale
    image_data = 255 * (image_data - data_min) / (data_max - data_min)
    image_data = np.minimum(image_data, 255)
    image_data = np.maximum(image_data, 0)
    image_data[np.isnan(image_data)] = 0
    return image_data


def remove_media_dirs(wandb_dir: str, media_dir_pattern: str = "run-*-*/files/media"):
    """
    Remove the media directories in the wandb run directories.
    """
    glob_pattern = os.path.join(wandb_dir, media_dir_pattern)
    media_dirs = glob.glob(glob_pattern)
    for media_dir in media_dirs:
        if os.path.isdir(media_dir):
            shutil.rmtree(media_dir)



import dataclasses
import datetime
import logging
from copy import copy
from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple, Union

import dacite
import torch
import xarray as xr
from torch import nn

from fme.ace.data_loading.batch_data import BatchData, PairedData, PrognosticState
from fme.ace.inference.derived_variables import compute_derived_quantities
from fme.ace.requirements import PrognosticStateDataRequirements
from fme.core.coordinates import HybridSigmaPressureCoordinate
from fme.core.corrector.corrector import CorrectorConfig
from fme.core.dataset.requirements import DataRequirements
from fme.core.dataset.utils import decode_timestep, encode_timestep
from fme.core.device import get_device
from fme.core.distributed import Distributed
from fme.core.generics.inference import PredictFunction
from fme.core.generics.optimization import OptimizationABC
from fme.core.generics.train_stepper import TrainOutputABC, TrainStepperABC
from fme.core.gridded_ops import GriddedOperations, LatLonOperations
from fme.core.loss import WeightedMappingLossConfig
from fme.core.normalizer import NormalizationConfig, StandardNormalizer
from fme.core.ocean import Ocean, OceanConfig
from fme.core.optimization import NullOptimization
from fme.core.packer import Packer
from fme.core.parameter_init import ParameterInitializationConfig
from fme.core.registry import CorrectorSelector, ModuleSelector
from fme.core.timing import GlobalTimer
from fme.core.typing_ import TensorDict, TensorMapping

DEFAULT_TIMESTEP = datetime.timedelta(hours=6)
DEFAULT_ENCODED_TIMESTEP = encode_timestep(DEFAULT_TIMESTEP)


class AtmosphericDeriveFn:
    def __init__(
        self,
        vertical_coordinate: HybridSigmaPressureCoordinate,
        timestep: datetime.timedelta,
    ):
        self.vertical_coordinate = vertical_coordinate.to(
            "cpu"
        )  # must be on cpu for multiprocessing fork context
        self.timestep = timestep

    def __call__(self, data: TensorMapping, forcing_data: TensorMapping) -> TensorDict:
        return compute_derived_quantities(
            dict(data),
            vertical_coordinate=self.vertical_coordinate.to(get_device()),
            timestep=self.timestep,
            forcing_data=dict(forcing_data),
        )


@dataclasses.dataclass
class SingleModuleStepperConfig:
    """
    Configuration for a single module stepper.

    Parameters:
        builder: The module builder.
        in_names: Names of input variables.
        out_names: Names of output variables.
        normalization: The normalization configuration.
        parameter_init: The parameter initialization configuration.
        ocean: The ocean configuration.
        loss: The loss configuration.
        corrector: The corrector configuration.
        next_step_forcing_names: Names of forcing variables for the next timestep.
        loss_normalization: The normalization configuration for the loss.
        residual_normalization: Optional alternative to configure loss normalization.
            If provided, it will be used for all *prognostic* variables in loss scaling.
    """

    builder: ModuleSelector
    in_names: List[str]
    out_names: List[str]
    normalization: NormalizationConfig
    parameter_init: ParameterInitializationConfig = dataclasses.field(
        default_factory=lambda: ParameterInitializationConfig()
    )
    ocean: Optional[OceanConfig] = None
    loss: WeightedMappingLossConfig = dataclasses.field(
        default_factory=lambda: WeightedMappingLossConfig()
    )
    corrector: Union[CorrectorConfig, CorrectorSelector] = dataclasses.field(
        default_factory=lambda: CorrectorConfig()
    )
    next_step_forcing_names: List[str] = dataclasses.field(default_factory=list)
    loss_normalization: Optional[NormalizationConfig] = None
    residual_normalization: Optional[NormalizationConfig] = None

    def __post_init__(self):
        for name in self.next_step_forcing_names:
            if name not in self.in_names:
                raise ValueError(
                    f"next_step_forcing_name '{name}' not in in_names: {self.in_names}"
                )
            if name in self.out_names:
                raise ValueError(
                    f"next_step_forcing_name is an output variable: '{name}'"
                )
        if (
            self.residual_normalization is not None
            and self.loss_normalization is not None
        ):
            raise ValueError(
                "Only one of residual_normalization, loss_normalization can "
                "be provided."
                "If residual_normalization is provided, it will be used for all "
                "*prognostic* variables in loss scalng. "
                "If loss_normalization is provided, it will be used for all variables "
                "in loss scaling."
            )

    @property
    def n_ic_timesteps(self) -> int:
        return 1

    def get_evaluation_window_data_requirements(
        self, n_forward_steps: int
    ) -> DataRequirements:
        return DataRequirements(
            names=self.all_names,
            n_timesteps=self._window_steps_required(n_forward_steps),
        )

    def get_prognostic_state_data_requirements(self) -> PrognosticStateDataRequirements:
        return PrognosticStateDataRequirements(
            names=self.prognostic_names,
            n_timesteps=self.n_ic_timesteps,
        )

    def get_forcing_window_data_requirements(
        self, n_forward_steps: int
    ) -> DataRequirements:
        if self.ocean is None:
            names = self.forcing_names
        else:
            names = list(set(self.forcing_names).union(self.ocean.forcing_names))

        return DataRequirements(
            names=names,
            n_timesteps=self._window_steps_required(n_forward_steps),
        )

    def _window_steps_required(self, n_forward_steps: int) -> int:
        return n_forward_steps + self.n_ic_timesteps

    def get_state(self):
        return dataclasses.asdict(self)

    def get_base_weights(self) -> Optional[List[Mapping[str, Any]]]:
        """
        If the model is being initialized from another model's weights for fine-tuning,
        returns those weights. Otherwise, returns None.

        The list mirrors the order of `modules` in the `SingleModuleStepper` class.
        """
        base_weights = self.parameter_init.get_base_weights()
        if base_weights is not None:
            return [base_weights]
        else:
            return None

    def get_stepper(
        self,
        img_shape: Tuple[int, int],
        gridded_operations: GriddedOperations,
        vertical_coordinate: HybridSigmaPressureCoordinate,
        timestep: datetime.timedelta,
    ):
        logging.info("Initializing stepper from provided config")
        derive_func = AtmosphericDeriveFn(vertical_coordinate, timestep)
        return SingleModuleStepper(
            config=self,
            img_shape=img_shape,
            gridded_operations=gridded_operations,
            vertical_coordinate=vertical_coordinate,
            timestep=timestep,
            derive_func=derive_func,
        )

    @classmethod
    def from_state(cls, state) -> "SingleModuleStepperConfig":
        state = cls.remove_deprecated_keys(state)
        return dacite.from_dict(
            data_class=cls, data=state, config=dacite.Config(strict=True)
        )

    @property
    def all_names(self):
        """Names of all variables required, including auxiliary ones."""
        extra_names = []
        if self.ocean is not None:
            extra_names.extend(self.ocean.forcing_names)
        all_names = list(set(self.in_names).union(self.out_names).union(extra_names))
        return all_names

    @property
    def normalize_names(self):
        """Names of variables which require normalization. I.e. inputs/outputs."""
        return list(set(self.in_names).union(self.out_names))

    @property
    def forcing_names(self) -> List[str]:
        """Names of variables which are inputs only."""
        return list(set(self.in_names) - set(self.out_names))

    @property
    def prognostic_names(self) -> List[str]:
        """Names of variables which both inputs and outputs."""
        return list(set(self.out_names).intersection(self.in_names))

    @property
    def diagnostic_names(self) -> List[str]:
        """Names of variables which both inputs and outputs."""
        return list(set(self.out_names).difference(self.in_names))

    @classmethod
    def remove_deprecated_keys(cls, state: Dict[str, Any]) -> Dict[str, Any]:
        _unsupported_key_defaults = {
            "conserve_dry_air": False,
            "optimization": None,
            "conservation_loss": {"dry_air_penalty": None},
        }
        state_copy = state.copy()
        for key, default in _unsupported_key_defaults.items():
            if key in state_copy:
                if state_copy[key] == default or state_copy[key] is None:
                    del state_copy[key]
                else:
                    raise ValueError(
                        f"The stepper config option {key} is deprecated and the setting"
                        f" provided, {state_copy[key]}, is no longer implemented. The "
                        "SingleModuleStepper being loaded from state cannot be run by "
                        "this version of the code."
                    )
        for normalization_key in [
            "normalization",
            "loss_normalization",
            "residual_normalization",
        ]:
            if state_copy.get(normalization_key) is not None:
                if "exclude_names" in state_copy[normalization_key]:
                    if state_copy[normalization_key]["exclude_names"] is not None:
                        raise ValueError(
                            "The exclude_names option in normalization config is no "
                            "longer supported, but excluded names were found in "
                            f"{normalization_key}."
                        )
                    else:
                        del state_copy[normalization_key]["exclude_names"]
        if "prescriber" in state_copy:
            # want to maintain backwards compatibility for this particular feature
            if state_copy["prescriber"] is not None:
                if state_copy.get("ocean") is not None:
                    raise ValueError("Cannot specify both prescriber and ocean.")
                state_copy["ocean"] = {
                    "surface_temperature_name": state_copy["prescriber"][
                        "prescribed_name"
                    ],
                    "ocean_fraction_name": state_copy["prescriber"]["mask_name"],
                    "interpolate": state_copy["prescriber"]["interpolate"],
                }
            del state_copy["prescriber"]
        return state_copy


@dataclasses.dataclass
class ExistingStepperConfig:
    """
    Configuration for an existing stepper. This is only designed to point to
    a serialized stepper checkpoint for loading, e.g., in the case of training
    resumption.

    Parameters:
        checkpoint_path: The path to the serialized checkpoint.
    """

    checkpoint_path: str

    def __post_init__(self):
        self._stepper_config = SingleModuleStepperConfig.from_state(
            self._load_checkpoint()["stepper"]["config"]
        )

    def _load_checkpoint(self) -> Mapping[str, Any]:
        return torch.load(self.checkpoint_path, map_location=get_device())

    def get_evaluation_window_data_requirements(
        self, n_forward_steps: int
    ) -> DataRequirements:
        return self._stepper_config.get_evaluation_window_data_requirements(
            n_forward_steps
        )

    def get_prognostic_state_data_requirements(self) -> PrognosticStateDataRequirements:
        return self._stepper_config.get_prognostic_state_data_requirements()

    def get_forcing_window_data_requirements(
        self, n_forward_steps: int
    ) -> DataRequirements:
        return self._stepper_config.get_forcing_window_data_requirements(
            n_forward_steps
        )

    def get_base_weights(self) -> Optional[List[Mapping[str, Any]]]:
        return self._stepper_config.get_base_weights()

    def get_stepper(self, img_shape, gridded_operations, vertical_coordinate, timestep):
        del img_shape  # unused
        logging.info(f"Initializing stepper from {self.checkpoint_path}")
        return SingleModuleStepper.from_state(self._load_checkpoint()["stepper"])


def _combine_normalizers(
    residual_normalizer: StandardNormalizer,
    model_normalizer: StandardNormalizer,
) -> StandardNormalizer:
    # Combine residual and model normalizers by overwriting the model normalizer
    # values that are present in residual normalizer. The residual normalizer
    # is assumed to have a subset of prognostic keys only.
    means, stds = copy(model_normalizer.means), copy(model_normalizer.stds)
    means.update(residual_normalizer.means)
    stds.update(residual_normalizer.stds)
    return StandardNormalizer(means=means, stds=stds)


def _prepend_timesteps(
    data: TensorMapping, timesteps: TensorMapping, time_dim: int = 1
) -> TensorDict:
    return {k: torch.cat([timesteps[k], v], dim=time_dim) for k, v in data.items()}


@dataclasses.dataclass
class TrainOutput(TrainOutputABC):
    metrics: TensorDict
    gen_data: TensorDict
    target_data: TensorDict
    time: xr.DataArray
    normalize: Callable[[TensorDict], TensorDict]
    derive_func: Callable[[TensorMapping, TensorMapping], TensorDict] = (
        lambda x, _: dict(x)
    )

    def remove_initial_condition(self, n_ic_timesteps: int) -> "TrainOutput":
        return TrainOutput(
            metrics=self.metrics,
            gen_data={k: v[:, n_ic_timesteps:] for k, v in self.gen_data.items()},
            target_data={k: v[:, n_ic_timesteps:] for k, v in self.target_data.items()},
            time=self.time[:, n_ic_timesteps:],
            normalize=self.normalize,
            derive_func=self.derive_func,
        )

    def copy(self) -> "TrainOutput":
        """Creates new dictionaries for the data but with the same tensors."""
        return TrainOutput(
            metrics=self.metrics,
            gen_data={k: v for k, v in self.gen_data.items()},
            target_data={k: v for k, v in self.target_data.items()},
            time=self.time,
            normalize=self.normalize,
            derive_func=self.derive_func,
        )

    def prepend_initial_condition(
        self,
        initial_condition: PrognosticState,
    ) -> "TrainOutput":
        """
        Prepends an initial condition to the existing stepped data.
        Assumes data are on the same device.
        For data windows > 0, the target IC is different from the generated IC
            and may be provided for correct calculation of tendencies.

        Args:
            initial_condition: Initial condition data.
        """
        batch_data = initial_condition.as_batch_data()
        return TrainOutput(
            metrics=self.metrics,
            gen_data=_prepend_timesteps(self.gen_data, batch_data.data),
            target_data=_prepend_timesteps(
                self.target_data,
                batch_data.data,
            ),
            time=xr.concat([batch_data.time, self.time], dim="time"),
            normalize=self.normalize,
            derive_func=self.derive_func,
        )

    def compute_derived_variables(
        self,
    ) -> "TrainOutput":
        gen_data = self.derive_func(self.gen_data, self.target_data)
        target_data = self.derive_func(self.target_data, self.target_data)
        return TrainOutput(
            metrics=self.metrics,
            gen_data=gen_data,
            target_data=target_data,
            time=self.time,
            normalize=self.normalize,
            derive_func=self.derive_func,
        )

    def get_metrics(self) -> TensorDict:
        return self.metrics


class SingleModuleStepper(
    TrainStepperABC[
        PrognosticState,
        BatchData,
        BatchData,
        PairedData,
        TrainOutput,
    ],
):
    """
    Stepper class for a single pytorch module.
    """

    TIME_DIM = 1
    CHANNEL_DIM = -3

    def __init__(
        self,
        config: SingleModuleStepperConfig,
        img_shape: Tuple[int, int],
        gridded_operations: GriddedOperations,
        vertical_coordinate: HybridSigmaPressureCoordinate,
        derive_func: Callable[[TensorMapping, TensorMapping], TensorDict],
        timestep: datetime.timedelta,
        init_weights: bool = True,
    ):
        """
        Args:
            config: The configuration.
            img_shape: Shape of domain as (n_lat, n_lon).
            gridded_operations: The gridded operations, e.g. for area weighting.
            vertical_coordinate: The vertical coordinate.
            derive_func: Function to compute derived variables.
            timestep: Timestep of the model.
            init_weights: Whether to initialize the weights. Should pass False if
                the weights are about to be overwritten by a checkpoint.
        """
        self._gridded_operations = gridded_operations  # stored for serializing
        n_in_channels = len(config.in_names)
        n_out_channels = len(config.out_names)
        self.in_packer = Packer(config.in_names)
        self.out_packer = Packer(config.out_names)
        self.normalizer = config.normalization.build(config.normalize_names)
        if config.ocean is not None:
            self.ocean: Optional[Ocean] = config.ocean.build(
                config.in_names, config.out_names, timestep
            )
        else:
            self.ocean = None
        self.module = config.builder.build(
            n_in_channels=n_in_channels,
            n_out_channels=n_out_channels,
            img_shape=img_shape,
        )
        module, self._l2_sp_tuning_regularizer = config.parameter_init.apply(
            self.module, init_weights=init_weights
        )
        self.module = module.to(get_device())
        self.derive_func = derive_func
        self._img_shape = img_shape
        self._config = config
        self._no_optimization = NullOptimization()

        dist = Distributed.get_instance()
        self._is_distributed = dist.is_distributed()
        self.module = dist.wrap_module(self.module)

        self._vertical_coordinates = vertical_coordinate.to(get_device())
        self._timestep = timestep

        self.loss_obj = config.loss.build(
            gridded_operations.area_weighted_mean, config.out_names, self.CHANNEL_DIM
        )

        self._corrector = config.corrector.build(
            gridded_operations=gridded_operations,
            vertical_coordinate=self.vertical_coordinate,
            timestep=timestep,
        )
        if config.loss_normalization is not None:
            self.loss_normalizer = config.loss_normalization.build(
                names=config.normalize_names
            )
        elif config.residual_normalization is not None:
            # Use residual norm for prognostic variables and input/output
            # normalizer for diagnostic variables in loss
            self.loss_normalizer = _combine_normalizers(
                residual_normalizer=config.residual_normalization.build(
                    config.prognostic_names
                ),
                model_normalizer=self.normalizer,
            )
        else:
            self.loss_normalizer = self.normalizer
        self.in_names = config.in_names
        self.out_names = config.out_names

        _1: PredictFunction[  # for type checking
            PrognosticState,
            BatchData,
            BatchData,
        ] = self.predict

        _2: PredictFunction[  # for type checking
            PrognosticState,
            BatchData,
            PairedData,
        ] = self.predict_paired

    @property
    def vertical_coordinate(self) -> HybridSigmaPressureCoordinate:
        return self._vertical_coordinates

    @property
    def timestep(self) -> datetime.timedelta:
        return self._timestep

    @property
    def surface_temperature_name(self) -> Optional[str]:
        if self._config.ocean is not None:
            return self._config.ocean.surface_temperature_name
        return None

    @property
    def ocean_fraction_name(self) -> Optional[str]:
        if self._config.ocean is not None:
            return self._config.ocean.ocean_fraction_name
        return None

    @property
    def effective_loss_scaling(self) -> TensorDict:
        """
        Effective loss scalings used to normalize outputs before computing loss.
        y_loss_normalized_i = (y_i - y_mean_i) / loss_scaling_i
        where loss_scaling_i = loss_normalizer_std_i / weight_i.
        """
        custom_weights = self._config.loss.weights
        loss_normalizer_stds = self.loss_normalizer.stds
        return {
            k: loss_normalizer_stds[k] / custom_weights.get(k, 1.0)
            for k in self._config.out_names
        }

    def replace_ocean(self, ocean: Ocean):
        """
        Replace the ocean model with a new one.

        Args:
            ocean: The new ocean model.
        """
        self.ocean = ocean

    @property
    def forcing_names(self) -> List[str]:
        """Names of variables which are inputs only."""
        return self._config.forcing_names

    @property
    def prognostic_names(self) -> List[str]:
        return sorted(
            list(set(self.out_packer.names).intersection(self.in_packer.names))
        )

    @property
    def diagnostic_names(self) -> List[str]:
        return sorted(list(set(self.out_packer.names).difference(self.in_packer.names)))

    @property
    def n_ic_timesteps(self) -> int:
        return 1

    @property
    def modules(self) -> nn.ModuleList:
        """
        Returns:
            A list of modules being trained.
        """
        return nn.ModuleList([self.module])

    def step(
        self,
        input: TensorMapping,
        next_step_forcing_data: TensorMapping,
    ) -> TensorDict:
        """
        Step the model forward one timestep given input data.

        Args:
            input: Mapping from variable name to tensor of shape
                [n_batch, n_lat, n_lon]. This data is used as input for `self.module`
                and is assumed to contain all input variables and be denormalized.
            next_step_forcing_data: Mapping from variable name to tensor of shape
                [n_batch, n_lat, n_lon]. This must contain the necessary forcing
                data at the output timestep for the ocean model and corrector.

        Returns:
            The denormalized output data at the next time step.
        """
        input_norm = self.normalizer.normalize(input)
        input_tensor = self.in_packer.pack(input_norm, axis=self.CHANNEL_DIM)
        output_tensor = self.module(input_tensor)
        output_norm = self.out_packer.unpack(output_tensor, axis=self.CHANNEL_DIM)
        output = self.normalizer.denormalize(output_norm)
        if self._corrector is not None:
            output = self._corrector(input, output, next_step_forcing_data)
        if self.ocean is not None:
            output = self.ocean(input, output, next_step_forcing_data)
        return output

    def _predict(
        self,
        initial_condition: TensorMapping,
        forcing_data: TensorMapping,
        n_forward_steps: int,
    ) -> TensorDict:
        """
        Predict multiple steps forward given initial condition and forcing data.

        Uses low-level inputs and does not compute derived variables, to separate
        concerns from the public `predict` method.

        Args:
            initial_condition: The initial condition, containing tensors of shape
                [n_batch, self.n_ic_timesteps, <horizontal_dims>].
            forcing_data: The forcing data, containing tensors of shape
                [n_batch, n_forward_steps + self.n_ic_timesteps, <horizontal_dims>].
            n_forward_steps: The number of forward steps to predict, corresponding
                to the data shapes of forcing_data.

        Returns:
            The output data at each timestep.
        """
        state = {
            k: initial_condition[k].squeeze(self.TIME_DIM) for k in initial_condition
        }
        ml_forcing_names = self._config.forcing_names
        output_list = []
        for step in range(n_forward_steps):
            ml_input_forcing = {
                k: (
                    forcing_data[k][:, step]
                    if k not in self._config.next_step_forcing_names
                    else forcing_data[k][:, step + 1]
                )
                for k in ml_forcing_names
            }
            next_step_forcing_data = {
                k: forcing_data[k][:, step + 1] for k in self._forcing_names()
            }
            input_data = {**state, **ml_input_forcing}
            state = self.step(input_data, next_step_forcing_data)
            output_list.append(state)
        output_timeseries = {}
        for name in state:
            output_timeseries[name] = torch.stack(
                [x[name] for x in output_list], dim=self.TIME_DIM
            )
        return output_timeseries

    def predict(
        self,
        initial_condition: PrognosticState,
        forcing: BatchData,
        compute_derived_variables: bool = False,
    ) -> Tuple[BatchData, PrognosticState]:
        """
        Predict multiple steps forward given initial condition and reference data.

        Args:
            initial_condition: Prognostic state data with tensors of shape
                [n_batch, self.n_ic_timesteps, <horizontal_dims>]. This data is assumed
                to contain all prognostic variables and be denormalized.
            forcing: Contains tensors of shape
                [n_batch, self.n_ic_timesteps + n_forward_steps, n_lat, n_lon]. This
                contains the forcing and ocean data for the initial condition and all
                subsequent timesteps.
            compute_derived_variables: Whether to compute derived variables for the
                prediction.

        Returns:
            A batch data containing the prediction and the prediction's final state
            which can be used as a new initial condition.
        """
        timer = GlobalTimer.get_instance()
        with timer.context("forward_prediction"):
            forcing_data = forcing.subset_names(self._forcing_names())
            initial_condition_state = initial_condition.as_batch_data()
            if initial_condition_state.time.shape[1] != self.n_ic_timesteps:
                raise ValueError(
                    f"Initial condition must have {self.n_ic_timesteps} timesteps, got "
                    f"{initial_condition_state.time.shape[1]}."
                )
            n_forward_steps = forcing_data.time.shape[1] - self.n_ic_timesteps
            output_timeseries = self._predict(
                initial_condition_state.data, forcing_data.data, n_forward_steps
            )
            data = BatchData.new_on_device(
                output_timeseries,
                forcing_data.time[:, self.n_ic_timesteps :],
                horizontal_dims=forcing_data.horizontal_dims,
            )
        if compute_derived_variables:
            with timer.context("compute_derived_variables"):
                data = (
                    data.prepend(initial_condition)
                    .compute_derived_variables(
                        derive_func=self.derive_func,
                        forcing_data=forcing_data,
                    )
                    .remove_initial_condition(self.n_ic_timesteps)
                )
        return data, data.get_end(self.prognostic_names, self.n_ic_timesteps)

    def predict_paired(
        self,
        initial_condition: PrognosticState,
        forcing: BatchData,
        compute_derived_variables: bool = False,
    ) -> Tuple[PairedData, PrognosticState]:
        """
        Predict multiple steps forward given initial condition and reference data.

        Args:
            initial_condition: Prognostic state data with tensors of shape
                [n_batch, self.n_ic_timesteps, <horizontal_dims>]. This data is assumed
                to contain all prognostic variables and be denormalized.
            forcing: Contains tensors of shape
                [n_batch, self.n_ic_timesteps + n_forward_steps, n_lat, n_lon]. This
                contains the forcing and ocean data for the initial condition and all
                subsequent timesteps.
            compute_derived_variables: Whether to compute derived variables for the
                prediction.

        Returns:
            A paired data containing the prediction paired with all forcing data at the
            same timesteps and the prediction's final state which can be used as a
            new initial condition.
        """
        prediction, new_initial_condition = self.predict(
            initial_condition, forcing, compute_derived_variables
        )
        return (
            PairedData.from_batch_data(
                prediction=prediction,
                target=self.get_forward_data(
                    forcing, compute_derived_variables=compute_derived_variables
                ),
            ),
            new_initial_condition,
        )

    def get_forward_data(
        self, data: BatchData, compute_derived_variables: bool = False
    ) -> BatchData:
        if compute_derived_variables:
            timer = GlobalTimer.get_instance()
            with timer.context("compute_derived_variables"):
                data = data.compute_derived_variables(
                    derive_func=self.derive_func,
                    forcing_data=data,
                )
        return data.remove_initial_condition(self.n_ic_timesteps)

    def _forcing_names(self) -> List[str]:
        if self.ocean is None:
            return self._config.forcing_names
        return list(set(self._config.forcing_names).union(self.ocean.forcing_names))

    def train_on_batch(
        self,
        data: BatchData,
        optimization: OptimizationABC,
        compute_derived_variables: bool = False,
    ) -> TrainOutput:
        """
        Step the model forward multiple steps on a batch of data.

        Args:
            data: The batch data where each tensor in data.data has shape
                [n_sample, n_forward_steps + self.n_ic_timesteps, <horizontal_dims>].
            optimization: The optimization class to use for updating the module.
                Use `NullOptimization` to disable training.
            compute_derived_variables: Whether to compute derived variables for the
                prediction and target data.

        Returns:
            The loss metrics, the generated data, the normalized generated data,
                and the normalized batch data.
        """
        time_dim = self.TIME_DIM

        loss = torch.tensor(0.0, device=get_device())
        metrics: Dict[str, float] = {}
        input_data = data.get_start(self.prognostic_names, self.n_ic_timesteps)

        optimization.set_mode(self.module)
        with optimization.autocast():
            # output from self.predict does not include initial condition
            output, _ = self.predict_paired(
                input_data,
                forcing=data,
            )
            gen_data = output.prediction
            target_data = output.target
            n_forward_steps = output.time.shape[1]

            # compute loss for each timestep
            for step in range(n_forward_steps):
                # Note: here we examine the loss for a single timestep,
                # not a single model call (which may contain multiple timesteps).
                gen_step = {k: v.select(time_dim, step) for k, v in gen_data.items()}
                target_step = {
                    k: v.select(time_dim, step) for k, v in target_data.items()
                }
                gen_norm_step = self.loss_normalizer.normalize(gen_step)
                target_norm_step = self.loss_normalizer.normalize(target_step)

                step_loss = self.loss_obj(gen_norm_step, target_norm_step)
                loss += step_loss
                metrics[f"loss_step_{step}"] = step_loss.detach()

        loss += self._l2_sp_tuning_regularizer()

        metrics["loss"] = loss.detach()
        optimization.step_weights(loss)

        stepped = TrainOutput(
            metrics=metrics,
            gen_data=dict(gen_data),
            target_data=dict(target_data),
            time=output.time,
            normalize=self.normalizer.normalize,
            derive_func=self.derive_func,
        )
        ic = data.get_start(
            set(data.data.keys()), self.n_ic_timesteps
        )  # full data and not just prognostic get prepended
        stepped = stepped.prepend_initial_condition(ic)
        if compute_derived_variables:
            stepped = stepped.compute_derived_variables()
        return stepped

    def get_state(self):
        """
        Returns:
            The state of the stepper.
        """
        return {
            "module": self.module.state_dict(),
            "normalizer": self.normalizer.get_state(),
            "img_shape": self._img_shape,
            "config": self._config.get_state(),
            "gridded_operations": self._gridded_operations.to_state(),
            "vertical_coordinate": self.vertical_coordinate.as_dict(),
            "encoded_timestep": encode_timestep(self.timestep),
            "loss_normalizer": self.loss_normalizer.get_state(),
        }

    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Load the state of the stepper.

        Args:
            state: The state to load.
        """
        if "module" in state:
            module = state["module"]
            if "module.device_buffer" in module:
                # for backwards compatibility with old checkpoints
                del module["module.device_buffer"]
            self.module.load_state_dict(module)

    @classmethod
    def from_state(cls, state) -> "SingleModuleStepper":
        """
        Load the state of the stepper.

        Args:
            state: The state to load.

        Returns:
            The stepper.
        """
        config = {**state["config"]}  # make a copy to avoid mutating input
        config["normalization"] = state["normalizer"]

        # for backwards compatibility with previous steppers created w/o
        # loss_normalization or residual_normalization
        loss_normalizer_state = state.get("loss_normalizer", state["normalizer"])
        config["loss_normalization"] = loss_normalizer_state

        # Overwrite the residual_normalization key if it exists, since the combined
        # loss scalings are saved in initial training as the loss_normalization
        config["residual_normalization"] = None

        if "area" in state:
            # backwards-compatibility, these older checkpoints are always lat-lon
            gridded_operations: GriddedOperations = LatLonOperations(state["area"])
        else:
            gridded_operations = GriddedOperations.from_state(
                state["gridded_operations"]
            )

        if "sigma_coordinates" in state:
            # for backwards compatibility with old checkpoints
            state["vertical_coordinate"] = state["sigma_coordinates"]

        vertical_coordinate = dacite.from_dict(
            data_class=HybridSigmaPressureCoordinate,
            data=state["vertical_coordinate"],
            config=dacite.Config(strict=True),
        )
        # for backwards compatibility with original ACE checkpoint which
        # serialized vertical coordinates as float64
        if vertical_coordinate.ak.dtype == torch.float64:
            vertical_coordinate.ak = vertical_coordinate.ak.to(dtype=torch.float32)
        if vertical_coordinate.bk.dtype == torch.float64:
            vertical_coordinate.bk = vertical_coordinate.bk.to(dtype=torch.float32)
        encoded_timestep = state.get("encoded_timestep", DEFAULT_ENCODED_TIMESTEP)
        timestep = decode_timestep(encoded_timestep)
        if "img_shape" in state:
            img_shape = state["img_shape"]
        else:
            # this is for backwards compatibility with old checkpoints
            for v in state["data_shapes"].values():
                img_shape = v[-2:]
                break
        derive_func = AtmosphericDeriveFn(vertical_coordinate, timestep)
        stepper = cls(
            config=SingleModuleStepperConfig.from_state(config),
            img_shape=img_shape,
            gridded_operations=gridded_operations,
            vertical_coordinate=vertical_coordinate,
            timestep=timestep,
            derive_func=derive_func,
            # don't need to initialize weights, we're about to load_state
            init_weights=False,
        )
        stepper.load_state(state)
        return stepper



from typing import Any, Dict, Mapping


def to_flat_dict(d: Mapping[str, Any]) -> Dict[str, Any]:
    """
    Converts any nested dictionaries to a flat version with
    the nested keys joined with a '.', e.g., {a: {b: 1}} ->
    {a.b: 1}.
    """
    new_flat = {}
    for k, v in d.items():
        if isinstance(v, dict):
            sub_d = to_flat_dict(v)
            for sk, sv in sub_d.items():
                new_flat[".".join([k, sk])] = sv
        else:
            new_flat[k] = v

    return new_flat


def to_nested_dict(d: Mapping[str, Any]) -> Dict[str, Any]:
    """
    Converts a flat dictionary with '.' joined keys back into
    a nested dictionary, e.g., {a.b: 1} -> {a: {b: 1}}.
    """
    new_config: Dict[str, Any] = {}

    for k, v in d.items():
        if "." in k:
            sub_keys = k.split(".")
            sub_d = new_config
            for sk in sub_keys[:-1]:
                sub_d = sub_d.setdefault(sk, {})
            sub_d[sub_keys[-1]] = v
        else:
            new_config[k] = v

    return new_config



import shutil
import tempfile
from pathlib import Path
from typing import Callable, Dict, Mapping, Optional, Sequence

import numpy as np
import torch
import xarray as xr

from fme.core.data_loading.data_typing import VariableMetadata


class PairedRestartWriter:
    """Wrapper around RestartWriter to handle generated and target data."""

    def __init__(
        self,
        path: str,
        is_restart_step: Callable[[int], bool],
        prognostic_names: Sequence[str],
        metadata: Mapping[str, VariableMetadata],
        coords: Mapping[str, np.ndarray],
    ):
        """
        Args:
            path: Path to the directory in which to save the file.
            is_restart_step: Function that returns True if the given timestep is a
                restart step.
            prognostic_names: Names of prognostic variables to write to the file.
            metadata: Metadata for each variable to be written to the file.
            coords: Coordinate data to be written to the file.
        """
        self._writer = RestartWriter(
            path, is_restart_step, prognostic_names, metadata, coords
        )

    def append_batch(
        self,
        target: Dict[str, torch.Tensor],
        prediction: Dict[str, torch.Tensor],
        start_timestep: int,
        batch_times: xr.DataArray,
    ):
        """
        Append a batch of data to the file.

        Args:
            target: Target data. Ignored for this data writer.
            prediction: Prediction data.
            start_timestep: Timestep (lead time dim) at which to start writing.
            batch_times: Time coordinates for each sample in the batch.
        """
        self._writer.append_batch(prediction, start_timestep, batch_times)

    def flush(self):
        pass


class RestartWriter:
    """
    Write raw prediction data to a netCDF file.
    """

    def __init__(
        self,
        path: str,
        is_restart_step: Callable[[int], bool],
        prognostic_names: Sequence[str],
        metadata: Mapping[str, VariableMetadata],
        coords: Mapping[str, np.ndarray],
    ):
        """
        Args:
            path: Path to the directory in which to save the file.
            is_restart_step: Function that returns True if the given timestep is a
                restart step.
            prognostic_names: Names of prognostic variables to write to the file.
            metadata: Metadata for each variable to be written to the file.
            coords: Coordinate data to be written to the file.
        """
        self.restart_filename = str(Path(path) / "restart.nc")
        self._save_names = prognostic_names
        self.metadata = metadata
        self.coords = coords
        self.is_restart_step = is_restart_step

    def append_batch(
        self,
        data: Dict[str, torch.Tensor],
        start_timestep: int,
        batch_times: xr.DataArray,
    ):
        """
        Append a batch of data to the file.

        Args:
            data: Data to write to the file.
            start_timestep: Timestep (lead time dim) at which to start writing.
            batch_times: Time coordinates for each sample in the batch.
        """
        n_times = data[list(data.keys())[0]].shape[1]
        restart_step = get_last_restart_step(
            start_timestep, n_times, self.is_restart_step
        )
        if restart_step is not None:
            window_step = restart_step - start_timestep
            self._write_restart_data(
                data, window_step, restart_step, batch_times[:, window_step]
            )

    def _write_restart_data(
        self,
        prediction: Dict[str, torch.Tensor],
        i_window: int,
        i_restart: int,
        restart_time: xr.DataArray,
    ):
        data_vars = {}
        for name in self._save_names:
            prediction_data = prediction[name][:, i_window, :, :].cpu().numpy()
            data_vars[name] = xr.DataArray(
                prediction_data, dims=["sample", "lat", "lon"]
            )
            if name in self.metadata:
                data_vars[name].attrs = {
                    "long_name": self.metadata[name].long_name,
                    "units": self.metadata[name].units,
                }
        data_vars["time"] = restart_time
        ds = xr.Dataset(
            data_vars=data_vars,
            coords=self.coords,
            attrs={"timestep": i_restart},
        )
        # first write to a temporary file so we have the old restart even if
        # the job ends or errors while writing the new one
        with tempfile.NamedTemporaryFile() as tmpfile:
            ds.to_netcdf(tmpfile.name)
            Path(self.restart_filename).unlink(missing_ok=True)
            shutil.copy(tmpfile.name, self.restart_filename)

    def flush(self):
        pass


def get_last_restart_step(
    start_step: int, n_steps: int, is_restart_step: Callable[[int], bool]
) -> Optional[int]:
    """
    Get the last restart step in a sequence of steps.

    Args:
        start_step: The first step in the sequence.
        n_steps: The number of steps in the sequence.
        is_restart_step: A function that returns True if a given step
            is a restart step.

    Returns:
        The last restart step in the sequence, or None if there is no restart step.
    """
    timesteps = list(range(n_steps))
    timesteps.reverse()  # we want the last step, so start at the end
    for i_time in timesteps:
        if is_restart_step(start_step + i_time):
            return start_step + i_time
    return None
